"""Entry point into upload command."""
import json
import os
import uuid
from datetime import datetime
from time import sleep

import backoff

from gencove.client import (  # noqa: I100
    APIClientError,
    APIClientTimeout,
    APIClientTooManyRequestsError,
    CustomEncoder,
)
from gencove.command.base import Command
from gencove.command.utils import is_valid_uuid
from gencove.constants import (
    FASTQ_MAP_EXTENSION,
    SampleAssignmentStatus,
    UPLOAD_PREFIX,
)
from gencove.exceptions import ValidationError
from gencove.utils import (
    batchify,
    get_regular_progress_bar,
    get_s3_client_refreshable,
)

from .constants import (
    ASSIGN_ERROR,
    FASTQ_EXTENSIONS,
    TMP_UPLOADS_WARNING,
    UploadStatuses,
)
from .exceptions import SampleSheetError, UploadError, UploadNotFound
from .multi_file_reader import MultiFileReader
from .utils import (
    get_filename_from_path,
    get_get_upload_details_retry_predicate,
    get_gncv_path,
    looks_like_url,
    parse_fastqs_map_file,
    seek_files_to_upload,
    upload_file,
    upload_multi_file,
)
from ..utils import is_valid_json
from ...constants import ASSIGN_BATCH_SIZE


# pylint: disable=too-many-instance-attributes
class Upload(Command):
    """Upload command executor."""

    # pylint: disable=too-many-arguments
    def __init__(self, source, destination, credentials, options, output, no_progress):
        super().__init__(credentials, options)
        self.source = source
        self.destination = destination
        self.project_id = options.project_id
        self.fastqs = []
        self.fastqs_map = {}
        self.upload_ids = set()
        self.output = output
        self.assigned_samples = []
        self.no_progress = no_progress
        self.metadata = options.metadata

    @staticmethod
    def generate_gncv_destination(fastq_source: str = "cli"):
        """Autogenerate gencove destination path."""
        return (
            f"{UPLOAD_PREFIX}{fastq_source}-"
            f"{datetime.utcnow().strftime('%Y%m%d%H%M%S')}-"
            f"{uuid.uuid4().hex}"
        )

    @staticmethod
    def map_paths_are_urls(fastq_map: dict) -> bool:
        """If all path values for fastq_map look like URLs,
        return True, otherwise False"""
        return all(looks_like_url(next(iter(path))) for path in fastq_map.values())

    def initialize(self):
        """Initialize upload command parameters from provided arguments."""
        self.echo_debug(f"Host is {self.options.host}")
        self.echo_warning(TMP_UPLOADS_WARNING)

        # fmt: off
        fastqs_source = "cli"
        if os.path.isfile(self.source) and self.source.endswith(FASTQ_MAP_EXTENSION):  # noqa: E501  # pylint: disable=line-too-long
            self.echo_debug("Scanning fastqs map file")
            self.fastqs_map = parse_fastqs_map_file(self.source)
            if self.map_paths_are_urls(self.fastqs_map):
                fastqs_source = "cli-url"
            self.echo_debug(f"got fastq pairs: {self.fastqs_map}")
        else:
            self.echo_debug("Seeking files to upload")
            self.fastqs = list(seek_files_to_upload(self.source))

        if not self.destination:
            self.destination = self.generate_gncv_destination(fastqs_source)
            self.echo_info(
                f"Files will be uploaded to: {self.destination}"
            )
        # fmt: on

        # Make sure there is just one trailing slash. Only exception is
        # UPLOAD_PREFIX itself, which can have two trailing slashes.
        if self.destination != UPLOAD_PREFIX:
            self.destination = self.destination.rstrip("/")
            self.destination += "/"

        if self.is_credentials_valid:
            self.login()

    def validate(self):
        """Validate command setup before execution.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        # fmt: off
        if self.destination and not self.destination.startswith(UPLOAD_PREFIX):
            self.echo_error(
                f"Invalid destination path. Must start with '{UPLOAD_PREFIX}'"
            )
            raise ValidationError("Bad configuration. Exiting.")
        # fmt: on

        if os.path.isfile(self.source) and not self.source.endswith(
            FASTQ_MAP_EXTENSION
        ):
            raise ValidationError(
                f"Source input ('{self.source}') is a file, "
                f"but must be a directory or a map file "
                f"('*{FASTQ_MAP_EXTENSION}'). Exiting."
            )

        if not self.fastqs and not self.fastqs_map:
            self.echo_error(
                "No FASTQ files found in the path. "
                f"Only following files are accepted: {FASTQ_EXTENSIONS}"
            )
            raise ValidationError("Bad configuration. Exiting.")

        if self.output and not self.project_id:
            raise ValidationError("--output cannot be used without --run-project-id")

        if self.metadata and not self.project_id:
            raise ValidationError("--metadata cannot be used without --run-project-id")

        if self.project_id and is_valid_uuid(self.project_id) is False:
            raise ValidationError("--run-project-id is not valid. Exiting.")

        # validate metadata json
        if self.metadata and is_valid_json(self.metadata) is False:
            raise ValidationError("--metadata is not valid JSON. Exiting.")

    def execute(self):
        """Upload fastq files from host system to Gencove cloud.

        If project id was provided, all fastq files will
        be assigned to this project, after upload.
        """
        s3_client = get_s3_client_refreshable(self.api_client.get_upload_credentials)

        try:
            if self.fastqs:
                self.upload_from_source(s3_client)
            elif self.fastqs_map:
                self.upload_from_map_file(s3_client)
        except UploadError:
            return

        self.echo_debug(f"Upload ids are now: {self.upload_ids}")
        if self.project_id:
            self.echo_debug("Cooling down period.")
            sleep(10)
            self.assign_uploads_to_project()
            if self.output:
                self.output_list()

    def upload_from_source(self, s3_client):
        """Upload command with <source> argument provided."""
        for file_path in self.fastqs:
            upload = self.upload_from_file_path(file_path, s3_client)
            if self.project_id and upload:
                self.upload_ids.add(upload.id)

        self.echo_info("All files were successfully uploaded.")

    def upload_from_map_file(self, s3_client):
        """Upload fastq files from a csv file."""
        for key, fastqs in self.fastqs_map.items():
            if all((looks_like_url(f) for f in fastqs)):
                upload = self.post_fastq_url(key, fastqs)
            else:
                upload = self.concatenate_and_upload_fastqs(key, fastqs, s3_client)

            if self.project_id and upload:
                self.upload_ids.add(upload.id)

        self.echo_info("All files were successfully processed.")

    def post_fastq_url(self, key, fastqs):
        """Post FASTQ URL to API endpoint"""
        client_id, r_notation = key
        gncv_path = self.destination + get_gncv_path(client_id, r_notation)
        self.echo_debug(f"Calculated gncv path: {gncv_path}")
        upload_url_details = self.api_client.import_fastqs_from_url(
            gncv_file_path=gncv_path,
            url=next(iter(fastqs)),
        )
        return upload_url_details

    def concatenate_and_upload_fastqs(self, key, fastqs, s3_client):
        """Upload fastqs parts as one file."""
        client_id, r_notation = key
        self.echo_debug(
            f"Uploading fastq. client_id={client_id} r_notation={r_notation}"
        )
        self.echo_debug(f"FASTQS: {fastqs}")

        gncv_path = self.destination + get_gncv_path(client_id, r_notation)
        self.echo_debug(f"Calculated gncv path: {gncv_path}")

        upload_details = self.get_upload_details(gncv_path)

        if (
            upload_details.last_status
            and upload_details.last_status.status == UploadStatuses.DONE.value
        ):
            self.echo_info(f"File was already uploaded: {gncv_path}")
            return upload_details

        self.echo_info(f"Uploading to {gncv_path}")
        upload_multi_file(
            s3_client,
            MultiFileReader(fastqs),
            upload_details.s3.bucket,
            upload_details.s3.object_name,
            self.no_progress,
        )
        return upload_details

    def upload_from_file_path(self, file_path, s3_client):
        """Prepare file and upload, if it wasn't uploaded yet.

        Args:
            file_path (str): a local system path to a file to be uploaded.
            s3_client (boto3 s3 client): instantiated boto3 S3 client.

        Returns:
            dict representing upload details
        """
        clean_file_path = get_filename_from_path(file_path, self.source)
        self.echo_debug(f"Uploading clean file path: {clean_file_path}")
        gncv_notated_path = self.destination + clean_file_path

        self.echo_info(f"Checking if file was already uploaded: {clean_file_path}")

        try:
            upload_details = self.get_upload_details(gncv_notated_path)
        except APIClientError as err:
            if err.status_code == 400:
                self.echo_info(err.message)
                raise UploadError  # pylint: disable=W0707
            raise err

        if (
            upload_details.last_status
            and upload_details.last_status.status == UploadStatuses.DONE.value
        ):
            self.echo_info(f"File was already uploaded: {clean_file_path}")
            return upload_details

        self.echo_info(f"Uploading {file_path} to {gncv_notated_path}")
        upload_file(
            s3_client=s3_client,
            file_name=file_path,
            bucket=upload_details.s3.bucket,
            object_name=upload_details.s3.object_name,
            no_progress=self.no_progress,
        )
        return upload_details

    @backoff.on_predicate(
        backoff.expo, get_get_upload_details_retry_predicate, max_tries=10
    )
    @backoff.on_exception(backoff.expo, APIClientError, max_tries=10)
    @backoff.on_exception(backoff.expo, APIClientTooManyRequestsError, max_tries=20)
    def get_upload_details(self, gncv_path):
        """Get upload details with retry for last status update."""
        return self.api_client.get_upload_details(gncv_path)

    def assign_uploads_to_project(self):
        """Assign uploads to a project and trigger a run."""
        self.echo_info(f"Assigning uploads to project {self.project_id}")

        try:
            samples = self.build_samples(self.upload_ids)
        except (UploadError, SampleSheetError, UploadNotFound):
            self.echo_warning(ASSIGN_ERROR.format(self.project_id, self.destination))
            return

        if not samples:
            self.echo_debug("No related samples were found")
            self.echo_warning(ASSIGN_ERROR.format(self.project_id, self.destination))
            return

        self.echo_debug(f"Sample sheet now is: {samples}")

        self.echo_debug(f"Assigning samples to project ({self.project_id})")

        assigned_count = 0
        if not self.no_progress:
            progress_bar = get_regular_progress_bar(len(samples), "Assigning: ")
            progress_bar.start()
        for samples_batch in batchify(samples, batch_size=ASSIGN_BATCH_SIZE):
            try:
                samples_batch_len = len(samples_batch)
                self.echo_debug(f"Assigning batch: {samples_batch_len}")
                metadata_api = None
                if self.metadata is not None:
                    metadata_api = json.loads(self.metadata)
                assigned_batch = self.api_client.add_samples_to_project(
                    samples_batch, self.project_id, metadata_api
                )
                if assigned_batch.uploads:
                    self.assigned_samples.extend(assigned_batch.uploads)
                assigned_count += samples_batch_len
                if not self.no_progress:
                    progress_bar.update(samples_batch_len)
                self.echo_debug(f"Total assigned: {assigned_count}")
            except APIClientError as err:
                self.echo_debug(err)
                self.echo_warning("There was an error assigning/running samples.")
                if assigned_count > 0:
                    self.echo_warning(
                        "Some of the samples were assigned. "
                        "Please use the Web UI to assign "
                        "the rest of the samples"
                    )
                else:
                    self.echo_warning(
                        ASSIGN_ERROR.format(self.project_id, self.destination)
                    )
                if not self.no_progress:
                    progress_bar.finish()
                return
        if not self.no_progress:
            progress_bar.finish()
        self.echo_info("Assigned all samples to a project")

    @backoff.on_exception(
        backoff.expo, (SampleSheetError, UploadNotFound), max_time=300
    )
    def build_samples(self, uploads):
        """Get samples for current uploads.

        Returns:
            list of dict: a list of samples for the uploads.
        """
        # make a copy of uploads so as not to change the input
        search_uploads = uploads.copy()
        samples = []
        for sample_sheet in self.sample_sheet_paginator():
            if not sample_sheet:
                self.echo_debug("Sample sheet returned empty.")
                raise UploadError

            for sample in sample_sheet:
                self.echo_debug(f"Checking sample: {sample}")
                add_it = False
                if sample.fastq and sample.fastq.r1:
                    if sample.fastq.r1.upload in search_uploads:
                        add_it = True
                        search_uploads.remove(sample.fastq.r1.upload)
                        self.echo_debug(
                            f"Found sample for upload r1: {sample.fastq.r1.upload}"
                        )
                    else:
                        self.echo_debug(
                            f"R1 upload not found. sample {sample} "
                            f"uploads {search_uploads}"
                        )
                        raise UploadNotFound
                if sample.fastq and sample.fastq.r2:
                    if sample.fastq.r2.upload in search_uploads:
                        add_it = True
                        search_uploads.remove(sample.fastq.r2.upload)
                        self.echo_debug(
                            f"Found sample for upload r2: {sample.fastq.r2.upload}"
                        )
                    else:
                        self.echo_debug(
                            f"R2 upload not found. sample {sample} "
                            f"uploads {search_uploads}"
                        )
                        raise UploadNotFound

                if add_it:
                    samples.append(sample)

        if search_uploads:
            self.echo_debug(f"Have uploads without samples: {search_uploads}")
            raise SampleSheetError

        return samples

    # duplicated in projects/run_prefix/main.py with the exception difference
    def sample_sheet_paginator(self):
        """Paginate over all sample sheets for the destination.

        Yields:
            paginated lists of samples
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get sample sheet page")
            try:
                resp = self.get_sample_sheet(next_link)
                yield resp.results
                next_link = resp.meta.next
                more = next_link is not None
            except APIClientError as err:
                self.echo_debug(err)
                raise UploadError  # pylint: disable=W0707

    @backoff.on_exception(
        backoff.expo,
        (APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    def get_sample_sheet(self, next_link=None):
        """Get samples by gncv path."""
        return self.api_client.get_sample_sheet(
            self.destination,
            SampleAssignmentStatus.UNASSIGNED.value,
            next_link,
        )

    def output_list(self):
        """Output JSON of assigning samples to a project."""
        self.echo_debug("Outputting JSON.")
        if self.output == "-":
            self.echo_data(
                json.dumps(self.assigned_samples, indent=4, cls=CustomEncoder)
            )
        else:
            dirname = os.path.dirname(self.output)
            if dirname and not os.path.exists(dirname):
                os.makedirs(dirname)
            with open(self.output, "w", encoding="utf-8") as json_file:
                json_file.write(
                    json.dumps(self.assigned_samples, indent=4, cls=CustomEncoder)
                )
            self.echo_info(f"Assigned samples response outputted to {self.output}")



"""Project create sample manifest shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, Optionals

from .main import CreateSampleManifest


@click.command("create-sample-manifest")
@click.argument("project_id", callback=validate_uuid)
@click.argument("sample_manifest")
@add_options(common_options)
def create_sample_manifest(  # pylint: disable=too-many-arguments
    project_id,
    sample_manifest,
    host,
    email,
    password,
    api_key,
):
    """Create a sample manifest in a project by uploading a CSV file.

    `PROJECT_ID`: Gencove project ID

    `SAMPLE_MANIFEST`: .csv file files that contains a table of information
        about a sequencing run
    """
    CreateSampleManifest(
        project_id,
        sample_manifest,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Utilities for processing file types."""


def get_line(file_type):
    """Build a list of relevant data to be printed.

    Args:
        file_type (dict): an object from file types response

    Returns:
        str: tab delimited list of relevant data to be printed.
    """
    return "\t".join(
        [
            file_type.key,
            file_type.description,
        ]
    )



"""Import existing samples shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid, validate_uuid_list
from gencove.constants import Credentials
from gencove.logger import echo_debug

from .constants import ImportExistingSamplesOptionals
from .main import ImportExistingSamples


@click.command("import-existing-samples")
@click.argument("project_id", callback=validate_uuid)
@click.option(
    "--source-project-id",
    "--project-id",
    required=False,
    help=(
        "Import all available samples, in succeeded or failed_qc state that have"
        " files, from source project."
        " Either --source-project-id or --source-sample-ids option must be provided,"
        " but not both."
    ),
)
@click.option(
    "--source-sample-ids",
    "--sample-ids",
    required=False,
    help=(
        "A comma separated list of source sample ids to import into the provided"
        " project."
        " Either --source-project-id or --source-sample-ids option must be provided,"
        " but not both."
    ),
    callback=validate_uuid_list,
)
@click.option(
    "--metadata-json",
    required=False,
    default=None,
    help="Add metadata to all samples that are to be imported into a project.",
)
@add_options(common_options)
def import_existing_project_samples(  # pylint: disable=too-many-arguments
    project_id,
    source_project_id,
    source_sample_ids,
    metadata_json,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Import existing samples to a project.

    `PROJECT_ID`: Gencove project ID

    Examples:

        Import samples from project:

            gencove projects import-existing-samples d9eaa54b-aaac-4b85-92b0-0b564be6d7db --source-project-id d8eb0bb5-29ee-44ed-b681-0fc05a557183

        Import samples:

            gencove projects import-existing-samples d9eaa54b-aaac-4b85-92b0-0b564be6d7db --sample-ids 59f5c1fd-cce0-4c4c-90e2-0b6c6c525d71,7edee497-12b5-4a1d-951f-34dc8dce1c1d

        Import samples with metadata:

            gencove projects import-existing-samples d9eaa54b-aaac-4b85-92b0-0b564be6d7db --sample-ids 59f5c1fd-cce0-4c4c-90e2-0b6c6c525d71,7edee497-12b5-4a1d-951f-34dc8dce1c1d --metadata-json='{"batch": "batch1"}'
    """  # noqa: E501
    echo_debug(f"Sample ids translation: {source_sample_ids}")
    ImportExistingSamples(
        project_id,
        source_project_id,
        source_sample_ids,
        Credentials(email=email, password=password, api_key=api_key),
        ImportExistingSamplesOptionals(host=host, metadata_json=metadata_json),
    ).run()



"""Commands to be executed from command line."""
# pylint: disable=E0012,C0330,R0913
import click

from .basespace_import.cli import basespace_import
from .basespace_list.cli import basespace_list


@click.group()
def projects():
    """BaseSpace projects management commands."""


projects.add_command(basespace_import)
projects.add_command(basespace_list)



"""Common code shared across data commands is stored here"""
import os
import subprocess  # nosec B404 (bandit subprocess import)
import sys
import uuid
from dataclasses import dataclass
from typing import List, Optional, Tuple

# pylint: disable=wrong-import-order
from gencove.exceptions import ValidationError
from gencove.models import ExplorerDataCredentials, OrganizationDetails, UserDetails

import boto3  # noqa: I100


@dataclass
class GencoveExplorerManager:  # pylint: disable=too-many-instance-attributes,too-many-public-methods # noqa: E501
    """Port of Explorer GencoveClient and related functionality"""

    user_id: str
    organization_id: str

    aws_session_credentials: Optional[ExplorerDataCredentials]

    # Constants ported from Gencove Explorer package
    # https://gitlab.com/gencove/platform/explorer-sdk/-/blob/main/gencove_explorer/constants.py  # noqa: E501 # pylint: disable=line-too-long
    # pylint: disable=invalid-name
    USERS: str = "users"
    ORG: str = "org"
    GENCOVE: str = "gencove"
    ME: str = "me"
    S3_PROTOCOL: str = "s3://"
    EXPLORER_SCHEME: str = "e://"
    DATA_BUCKET: str = "gencove-explorer-data"
    USER_DIR: str = "files"
    DATA_DIR: str = USER_DIR
    SHARED_DIR: str = "shared"
    SCRATCH_DIR: str = "scratch"
    USERS_DIR: str = "users"
    TMP_DIR: str = f"{DATA_DIR}/tmp"  # nosec: B108

    # pylint: enable=invalid-name

    def __post_init__(self):
        # pylint: disable=invalid-name
        self.NAMESPACES: dict = {
            self.USERS: self.data_s3_prefix,
            self.ORG: self.data_org_s3_prefix,
            self.GENCOVE: self.data_gencove_s3_prefix,
        }
        self.NAMESPACE_KEYS: Tuple = tuple(self.NAMESPACES.keys())
        # pylint: enable=invalid-name

    @property
    def bucket_name(self) -> str:
        """Construct bucket name based on short form of org ID"""
        organization_id_short = self.organization_id.replace("-", "")[:12]
        return f"gencove-explorer-{organization_id_short}"

    @property
    def aws_env(self) -> dict:
        """Dict containing AWS credentials"""
        aws_env = {}
        if self.aws_session_credentials:
            aws_env = {
                "AWS_ACCESS_KEY_ID": self.aws_session_credentials.access_key,
                "AWS_SECRET_ACCESS_KEY": self.aws_session_credentials.secret_key,
                "AWS_SESSION_TOKEN": self.aws_session_credentials.token,
                "AWS_DEFAULT_REGION": self.aws_session_credentials.region_name,
                "AWS_REGION": self.aws_session_credentials.region_name,
            }
        return aws_env

    @property
    def users_prefix(self) -> str:
        """S3 prefix for all users"""
        return f"{self.USERS_DIR}"

    @property
    def base_prefix(self) -> str:
        """Base S3 prefix for all objects to be uploaded/retrieved for user.

        e.g. user/1111/
        """
        return f"{self.users_prefix}/{self.user_id}"

    @property
    def shared_prefix(self) -> str:
        """Base S3 prefix for all objects to be uploaded/retrieved for organization."""
        return self.SHARED_DIR

    @property
    def data_org_prefix(self) -> str:
        """Base S3 prefix for all data to be uploaded/retreived for organization."""
        return f"{self.SHARED_DIR}/{self.DATA_DIR}"

    @property
    def data_prefix(self) -> str:
        """Base S3 prefix for all data to be uploaded/retreived for user."""
        return f"{self.base_prefix}/{self.DATA_DIR}"

    @property
    def user_prefix(self) -> str:
        """Base S3 prefix for user dir"""
        return f"{self.base_prefix}/{self.USER_DIR}"

    @property
    def tmp_prefix(self) -> str:
        """Base S3 prefix for temp user dir"""
        return f"{self.base_prefix}/{self.TMP_DIR}"

    @property
    def tmp_org_prefix(self) -> str:
        """S3 prefix for temp shared dir"""
        return f"{self.shared_prefix}/{self.TMP_DIR}"

    @property
    def scratch_prefix(self) -> str:
        """Base S3 prefix for scratch dir"""
        return f"{self.base_prefix}/{self.SCRATCH_DIR}"

    @property
    def user_scratch_s3_prefix(self) -> str:
        """S3 prefix for user scratch dir"""
        return f"{self.S3_PROTOCOL}{self.bucket_name}/{self.scratch_prefix}"

    @property
    def shared_org_s3_prefix(self) -> str:
        """S3 prefix for org dir"""
        return f"{self.S3_PROTOCOL}{self.bucket_name}/{self.shared_prefix}"

    @property
    def data_gencove_s3_prefix(self) -> str:
        """S3 prefix for gencove data dir"""
        return f"{self.S3_PROTOCOL}{self.DATA_BUCKET}/{self.DATA_DIR}"

    @property
    def data_org_s3_prefix(self) -> str:
        """S3 prefix for org data dir"""
        return f"{self.S3_PROTOCOL}{self.bucket_name}/{self.data_org_prefix}"

    @property
    def data_s3_prefix(self) -> str:
        """S3 prefix for user data dir"""
        return f"{self.S3_PROTOCOL}{self.bucket_name}/{self.data_prefix}"

    @property
    def users_s3_prefix(self) -> str:
        """Prefix for users dir"""
        return f"{self.S3_PROTOCOL}{self.bucket_name}/{self.users_prefix}"

    def run_s3_command(self, s3_command: List[str]) -> None:
        """Run AWS S3 command while handling AWS CLI non-zero exits gracefully

        Args:
            s3_command (List[str]): List of arguments to pass to AWS CLI
        """
        command = ["aws", "s3"] + s3_command
        env = self.aws_env.copy()
        env["PATH"] = os.environ["PATH"]
        try:
            subprocess.run(  # nosec B603 (execution of untrusted input)
                command,
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr,
                env=env,
                check=True,
            )
        except subprocess.CalledProcessError as err:
            if err.stderr:
                self.echo_error(err.stderr.decode())  # pylint: disable=no-member
            sys.exit(err.returncode)

    def uri_ok(self, path: Optional[str]) -> bool:
        """Tests if supplied path is valid

        Args:
            path (Optional[str]): Path to test

        Returns:
            True if path valid, False otherwise
        """
        return path is not None and path.startswith(self.EXPLORER_SCHEME)

    def translate_path_to_s3_path(self, path: Optional[str]) -> Optional[str]:
        """Accepts any input path and converts `e://` path to `s3://` path if
        input is an `e://` path

        Supported paths:
        e://gencove/...
        e://org/...
        e://users/<user-id>/...
        e://users/me/...

        Args:
            path (Optional[str]): Path to convert

        Returns:
            Converted path
        """
        if path is None:
            return None
        if path.startswith(self.EXPLORER_SCHEME):
            path_noprefix = path[len(self.EXPLORER_SCHEME) :]  # noqa: E203
            path_noprefix_split = path_noprefix.split("/")
            namespace = path_noprefix_split[0]
            if namespace not in self.NAMESPACES:
                raise ValueError(
                    f"Invalid namespace '{namespace}'. Valid namespaces are: "
                    f"{', '.join(self.NAMESPACE_KEYS)}"
                )
            prefix_s3 = self.NAMESPACES[namespace]
            path_remainder = "/".join(path_noprefix_split[1:])
            if namespace == self.USERS:
                if len(path_noprefix_split) <= 1:
                    raise ValueError(f"A user id (or '{self.ME}') must be specified")
                user_id = path_noprefix_split[1]
                if user_id != self.ME:
                    try:
                        uuid.UUID(user_id)
                    except ValueError:
                        raise ValueError(  # pylint: disable=raise-missing-from
                            f"User id '{user_id}' is not a valid UUID (or '{self.ME}')"
                        )
                    prefix_s3 = self.users_s3_prefix + f"/{user_id}/{self.USER_DIR}"
                path_remainder = "/".join(path_noprefix_split[2:])
            path = f"{prefix_s3}/{path_remainder}"
        return path

    def list_users(self):
        """List e:// user dir"""
        user_prefix = f"{self.S3_PROTOCOL}{self.bucket_name}/{self.USERS_DIR}/"
        command = ["aws", "s3", "ls", user_prefix]
        env = self.aws_env.copy()
        env["PATH"] = os.environ["PATH"]
        try:
            subprocess.run(  # nosec B603 (execution of untrusted input)
                command,
                stdin=sys.stdin,
                stdout=sys.stdout,
                stderr=sys.stderr,
                env=env,
                check=True,
            )
        except subprocess.CalledProcessError as err:
            sys.stderr.write(f"Error listing users: {err}\n")
            sys.exit(err.returncode)

    def execute_aws_s3_path(
        self,
        cmd: str,
        path: str,
        args: List[str],
    ) -> List[str]:
        """Executes the respective `aws s3` single-path commands with `e://` paths
        translated to `s3://` paths

        Args:
            cmd (str): AWS S3 command to execute
            path (str): Path to execute command against
            args (List[str]): List of additional args to forward to AWS CLI

        Returns:
            List of args passed to AWS CLI
        """
        if not self.uri_ok(path):
            raise ValueError(f"Path {path} does not start with {self.EXPLORER_SCHEME}")
        s3_command = [
            cmd,
            self.translate_path_to_s3_path(path),
            *args,
        ]
        self.run_s3_command(s3_command)
        return s3_command

    def execute_aws_s3_src_dst(
        self, cmd: str, source: str, destination: str, args: List[str]
    ) -> List[str]:
        """Executes the respective `aws s3` dual-path (source-to-destination)
        commands with e://` paths translated to `s3://` paths

        Args:
            cmd (str): AWS S3 command to execute
            source: Source path for S3 command
            destination: Destination path for S3 command
            args: List of additional args to forward to AWS CLI

        Returns:
            List of args passed to AWS CLI
        """
        if not self.uri_ok(source) and not self.uri_ok(destination):
            raise ValueError(
                f"At least one of source or destination must start with "
                f"{self.EXPLORER_SCHEME}"
            )
        s3_command = [
            cmd,
            self.translate_path_to_s3_path(source),
            self.translate_path_to_s3_path(destination),
            *args,
        ]
        self.run_s3_command(s3_command)
        return s3_command

    def thread_safe_client(self, service_name, *args, **kwargs):
        """Thread safe boto client with explorer session credentials.

        Args:
            service_name (str): Name of the service.

        Returns:
            BotoClient: Thread safe client.
        """
        # Thread-safe, as per:
        # https://boto3.amazonaws.com/v1/documentation/api/1.17.90/guide/clients.html#multithreading-or-multiprocessing-with-clients
        if self.aws_session_credentials:
            boto_session = boto3.Session(
                aws_access_key_id=self.aws_session_credentials.access_key,
                aws_secret_access_key=self.aws_session_credentials.secret_key,
                aws_session_token=self.aws_session_credentials.token,
                region_name=self.aws_session_credentials.region_name,
            )
        else:
            boto_session = boto3.Session()
        return boto_session.client(service_name, *args, **kwargs)

    def list_s3_objects(self, path: str):
        """List S3 objects in given path.

        Args:
            path (str): Path to s3 objects.

        Returns:
            PaginatedResponse: List of objects paginated.
        """
        s3_client = self.thread_safe_client("s3")
        s3_path = self.translate_path_to_s3_path(path)
        bucket, prefix = s3_path.lstrip("s3://").split("/", 1)
        paginated_response = s3_client.get_paginator("list_objects_v2").paginate(
            Bucket=bucket,
            Prefix=prefix,
            PaginationConfig={
                "PageSize": 1_000,
            },
        )
        return paginated_response


def validate_explorer_user_data(user: UserDetails, organization: OrganizationDetails):
    """Validate user and organization data

    Args:
        user (UserDetails): User details model
        organization (OrganizationDetails): Organization details model
    """
    if not user.explorer_enabled:
        raise ValidationError(
            "Explorer is not enabled on your user account, quitting. "
            "Please reach out to your organization owner to inquire "
            "about Gencove Explorer."
        )

    if not user or not organization:
        raise ValidationError("Could not retrieve user details, quitting.")


def request_is_from_explorer() -> bool:
    """
    Detects whether user is executing code from an Explorer environment.
    If check fails, we assume user is not in environment.

    Returns:
        bool: True if user in Explorer environment, False otherwise
    """
    try:
        user_id = os.environ.get("GENCOVE_USER_ID", None)
        if user_id is None:
            return False
        user_id_dashes = str(uuid.UUID(user_id))
        expected_role_name_instance = f"explorer-user-{user_id}-role"
        expected_role_name_cluster = f"explorer-{user_id_dashes}-ecs_task_role"
        client = boto3.client("sts")
        response = client.get_caller_identity()
        if (
            expected_role_name_instance in response["Arn"]
            or expected_role_name_cluster in response["Arn"]
        ):
            return True
    except Exception:  # noqa pylint: disable=broad-except
        return False
    return False



"""Utils for upload command."""
import csv
import os
import platform
import re
import urllib.parse
from collections import defaultdict
from urllib.parse import urlparse

from boto3.s3.transfer import TransferConfig

from botocore.exceptions import ClientError

from gencove.exceptions import ValidationError
from gencove.logger import echo_debug, echo_info
from gencove.utils import CHUNK_SIZE, get_progress_bar

from .constants import (
    FASTQ_EXTENSIONS,
    FastQ,
    PATH_TEMPLATE,
    PathTemplateParts,
    R_NOTATION_MAP,
)


def upload_file(
    s3_client, file_name, bucket, object_name=None, no_progress=False
):  # noqa: D413
    """Upload a file to an S3 bucket.

    Args:
        s3_client: Boto s3 client.
        file_name (str): File to upload.
        bucket (str): Bucket to upload to.
        object_name (str): S3 object name.
            If not specified then file_name is used

    Returns:
        True if file was uploaded, else False
    """
    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_name

    # Upload the file
    try:
        # Set desired multipart threshold value of 5GB
        config = TransferConfig(
            multipart_threshold=CHUNK_SIZE,
            multipart_chunksize=CHUNK_SIZE,
            use_threads=True,
            max_concurrency=10,
        )
        if not no_progress:
            progress_bar = get_progress_bar(os.path.getsize(file_name), "Uploading: ")
            progress_bar.start()
        s3_client.upload_file(
            file_name,
            bucket,
            object_name,
            Config=config,
            Callback=_progress_bar_update(progress_bar) if not no_progress else None,
        )
        if not no_progress:
            progress_bar.finish()
    except ClientError as err:
        echo_info(f"Failed to upload file {file_name}: {err}")
        return False
    return True


def upload_multi_file(
    s3_client,
    file_obj,
    bucket,
    object_name=None,  # pylint: disable=E0012,C0330
    no_progress=False,
):  # noqa: D413
    """Upload a file to an S3 bucket.

    Args:
        s3_client: Boto s3 client.
        file_obj (MultiFileReader): File-like object with read() and
            __iter__ methods
        bucket (str): Bucket to upload to.
        object_name (str): S3 object name.
            If not specified then file_name is used

    Returns:
        True if file was uploaded, else False
    """
    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_obj.name

    # Upload the file
    try:
        # Set desired multipart threshold value of 5GB
        config = TransferConfig(
            multipart_threshold=CHUNK_SIZE,
            multipart_chunksize=CHUNK_SIZE,
            use_threads=True,
            max_concurrency=10,
        )

        if not no_progress:
            progress_bar = get_progress_bar(file_obj.get_size(), "Uploading: ")
            progress_bar.start()
        s3_client.upload_fileobj(
            file_obj,
            bucket,
            object_name,
            Config=config,
            Callback=_progress_bar_update(progress_bar) if not no_progress else None,
        )
        if not no_progress:
            progress_bar.finish()
    except ClientError as err:
        echo_info(f"Failed to upload file {file_obj.name}: {err}")
        return False
    return True


def _progress_bar_update(pbar):  # noqa: D413
    """Update progress bar manually.

    Helper method for S3 Transfer,
    which needs a callback to update the progressbar.

    Args:
        pbar: progressbar.ProgressBar instance

    Returns:
        a function that in turn accepts chunk that is used to update the
        progressbar.
    """

    # noqa: D202
    def _update_pbar(chunk_uploaded_in_bytes):
        pbar.update(pbar.value + chunk_uploaded_in_bytes)

    return _update_pbar


def seek_files_to_upload(path, path_root=""):
    """Generate a list of valid fastq files."""
    for root, dirs, files in os.walk(path):
        files.sort()
        for file in files:
            file_path = os.path.join(path_root, root, file)

            if file_path.lower().endswith(FASTQ_EXTENSIONS):
                echo_debug(f"Found file to upload: {file_path}")
                yield file_path

        dirs.sort()
        for folder in dirs:
            seek_files_to_upload(folder, root)


def get_get_upload_details_retry_predicate(resp):
    """Triggers retry if upload details came back without last status."""
    return not resp.last_status


def get_filename_from_path(full_path, source):
    """Cross OS get file name utility."""
    relpath = os.path.relpath(os.path.normpath(full_path), start=source)
    if platform.system() == "Windows":
        return relpath.replace("\\", "/")
    return relpath


def _validate_fastq(fastq):
    """Validate fastq object.

    Args:
        fastq (FastQ): fastq object
    Returns:
        None if all good
    Raises:
        ValidationError if fastq is not valid
    """
    if looks_like_url(fastq.path):
        if not valid_fastq_file_name_in_url(fastq.path):
            raise ValidationError(
                f"Could not determine FASTQ file name from supplied URL {fastq.path}"
            )
    else:
        if not fastq.path.lower().endswith(FASTQ_EXTENSIONS):
            echo_info(f"Unsupported file type found: {fastq.path}")
            raise ValidationError(f"Bad file extension in path: {fastq.path}")
        if not os.path.exists(fastq.path):
            echo_info(f"Path does not exist: {fastq.path}")
            raise ValidationError(f"Could not find: {fastq.path}")

    if "_" in fastq.client_id:
        echo_info("Underscore is not allowed in client id")
        raise ValidationError("Underscore is not allowed in client id")
    if fastq.r_notation not in R_NOTATION_MAP:
        echo_info(f"Wrong R notation: {fastq.r_notation}")
        echo_info(f"Valid R notations are: {R_NOTATION_MAP.keys()}")
        raise ValidationError(f"Wrong R notation: {fastq.r_notation}")


def _validate_header(header):
    header_columns = [header_item.strip().lower() for header_item in header.values()]
    for column in FastQ.__fields__:
        if column not in header_columns:
            raise ValidationError(
                f"Unexpected CSV header. Expected: {', '.join(FastQ.__fields__)}"
            )


def _validate_parsed_map(fastqs: dict):
    """Validates contents of fastqs dict to ensure path
    column contains homogenous values, i.e. all local paths
     or all URLs"""
    paths = [next(iter(x)) for x in fastqs.values()]
    all_urls = all(looks_like_url(p) for p in paths)
    all_paths = all(not looks_like_url(p) for p in paths)
    if not all_urls and not all_paths:
        raise ValidationError(
            "Detected both URLs and file paths in 'path' column. "
            "Please only supply one type of path (URL or local path)"
            " in the map file."
        )


def parse_fastqs_map_file(fastqs_map_path):
    """Parse fastq map file.

    Map file has to have following columns/headers:
        client_id, r_notation, path

    Note this map file also supports URLs under the path column

    Example fastqs map file:
        client_id,r_notation,path
        sample1,R1,dir1/sample1_L001_R1.fastq.gz
        sample1,R1,dir1/sample1_L002_R1.fastq.gz
        sample2,R2,dir2/sample1_L001_R2.fastq.gz

    Args:
        fastqs_map_path (str): path to CSV file

    Returns:
        defaultdict: map of fastq file to samples
            {
                (<client_id>, <r_notation>): [path1, path2, ...],
            }
    """
    fastqs = defaultdict(list)
    with open(fastqs_map_path, encoding="utf-8") as fastqs_file:
        reader = csv.DictReader(fastqs_file, fieldnames=FastQ.__fields__)
        # read headers row
        header = next(reader)
        _validate_header(header)
        for row in reader:
            fastq = FastQ(**row)
            _validate_fastq(fastq)
            fastqs[(fastq.client_id, R_NOTATION_MAP[fastq.r_notation])].append(
                fastq.path
            )
    _validate_parsed_map(fastqs)
    return fastqs


def get_gncv_path(client_id, r_notation):
    """Build gncv upload path.

    Args:
        client_id (str): sample client id
        r_notation (str): upload R1/2 notation

    Returns:
        str: gncv upload path
    """
    return PATH_TEMPLATE.format(
        **{
            PathTemplateParts.client_id.value: client_id,
            PathTemplateParts.r_notation.value: r_notation,
        }
    )


def looks_like_url(value):
    """Detects if input value appears to be a URL

    Args:
        value (str): value to test

    Returns:
        bool: True if URL-like, False otherwise
    """
    try:
        result = urlparse(value)
        if result.scheme in ["http", "https"]:
            if result.netloc:
                return True
    except ValueError:
        return False
    return False


def valid_fastq_file_name_in_url(value):
    """Attempt to detect FASTQ file name in URL"""
    path = urllib.parse.urlparse(value).path
    fastq_extensions_regex = [x.replace(".", r"\.") for x in FASTQ_EXTENSIONS]
    re_fastq_extensions = "|".join(fastq_extensions_regex)
    re_pattern = rf"/([^/]*?({re_fastq_extensions}))$"
    match = re.search(re_pattern, path)
    if match:
        return True
    return False



"""Test s3_imports autoimports list command."""
# pylint: disable=wrong-import-order, import-error

import pytest
from click.testing import CliRunner  # noqa: I100,I201
from gencove.client import APIClient, APIClientTimeout  # noqa: I201
from gencove.command.s3_imports.autoimports.autoimport_list.cli import (
    autoimport_list,
)
from gencove.models import S3ProjectImport
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/s3_imports/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
        ],
        "before_record_response": [
            filter_jwt,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_autoimport_list_empty(mocker, credentials):
    """Test user has no jobs."""
    runner = CliRunner()
    mocked_list_s3_autoimport_jobs = mocker.patch.object(
        APIClient,
        "list_s3_autoimport_jobs",
        return_value=S3ProjectImport(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(autoimport_list, credentials)
    assert res.exit_code == 0
    mocked_list_s3_autoimport_jobs.assert_called_once()
    assert "" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_autoimport_list_uploads_slow_response_retry(mocker, credentials):
    """Test list jobs slow response retry."""
    runner = CliRunner()
    mocked_list_s3_autoimport_jobs = mocker.patch.object(
        APIClient,
        "list_s3_autoimport_jobs",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )
    res = runner.invoke(autoimport_list, credentials)
    assert res.exit_code == 1
    assert mocked_list_s3_autoimport_jobs.call_count == 2
    assert (
        "\n".join(
            [
                "ERROR: There was an error listing S3 autoimport jobs.",
                "ERROR: Could not connect to the api server",
                "Aborted!\n",
            ]
        )
        == res.output
    )


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_autoimport_list(mocker, credentials):
    """Test list autoimport jobs being outputed to the shell."""
    runner = CliRunner()
    autoimport_response = {
        "meta": {"count": 2, "next": None, "previous": None},
        "results": [
            {
                "id": "41fd0397-62b0-4ef9-992f-423435b5d5ef",
                "project_id": "290e0c5a-c87e-474e-8b32-fe56fc54cc4d",
                "s3_uri": "s3://bucket/project-1/samples-1/",
                "metadata": None,
            },
            {
                "id": "0f60ab5e-a34f-4afc-a428-66f81890565f",
                "project_id": "290e0c5a-c87e-474e-8b32-fe56fc54cc4d",
                "s3_uri": "s3://bucket/project-1/samples-2/",
                "metadata": {},
            },
        ],
    }
    mocked_list_s3_autoimport_jobs = mocker.patch.object(
        APIClient,
        "list_s3_autoimport_jobs",
        return_value=S3ProjectImport(**autoimport_response),
    )
    res = runner.invoke(autoimport_list, credentials)

    assert res.exit_code == 0

    mocked_list_s3_autoimport_jobs.assert_called()
    autoimport_jobs = autoimport_response["results"]
    jobs = "\n".join(
        [
            "\t".join(
                [
                    job["id"],
                    job["project_id"],
                    job["s3_uri"],
                ]
            )
            for job in autoimport_jobs
        ]
    )
    assert f"{jobs}\n" == res.output



"""Create project's batch executor."""
from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.projects.list_batches.utils import get_line
from gencove.exceptions import ValidationError


class CreateBatch(Command):
    """Create project's batch executor."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        project_id,
        batch_type,
        batch_name,
        sample_ids,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.batch_type = batch_type
        self.batch_name = batch_name
        self.sample_ids = sample_ids

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if not self.batch_type:
            raise ValidationError("You must provide value for --batch-type. Exiting.")

        if not self.batch_name:
            raise ValidationError("You must provide value for --batch-name. Exiting.")

    # no retry for timeouts in order to avoid duplicate heavy operations on
    # the backend
    def execute(self):
        """Make a request to create a batch for given project."""
        self.echo_debug(
            f"Creating batch for project {self.project_id} and "
            f"batch key {self.batch_name}"
        )

        try:
            created_batches_details = self.api_client.create_project_batch(
                project_id=self.project_id,
                batch_type=self.batch_type,
                batch_name=self.batch_name,
                sample_ids=self.sample_ids,
            )
            self.echo_debug(created_batches_details)
            for batch in created_batches_details.results:
                self.echo_data(get_line(batch))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error creating project batches.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(f"Project {self.project_id} does not exist.")
            else:
                raise



"""Test basespace projects list command."""
# pylint: disable=wrong-import-order
import io
import sys
from datetime import datetime, timedelta

from click import echo
from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
    APIClientTimeout,
)  # noqa: I100
from gencove.command.basespace.projects.cli import basespace_list
from gencove.models import (
    BaseSpaceProject,
    BaseSpaceProjectDetail,
)


def test_basespace_list__empty(mocker):
    """Test user has no BaseSpace projects."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_list_basespace_projects = mocker.patch.object(
        APIClient,
        "list_basespace_projects",
        return_value=BaseSpaceProject(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(basespace_list, ["--email", "foo@bar.com", "--password", "123"])
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_list_basespace_projects.assert_called_once()
    assert "" in res.output


MOCKED_BASESPACE_PROJECTS = dict(
    meta=dict(next=None),
    results=[
        {
            "basespace_id": "1234",
            "basespace_name": "test\tproject",
            "basespace_date_created": (
                datetime.utcnow() - timedelta(days=7)
            ).isoformat(),
        }
    ],
)

# API responses may return new keys and values eventually
MOCKED_BASESPACE_PROJECTS_WITH_UNEXPECTED_KEYS = dict(
    meta=dict(next=None),
    results=[
        {
            "basespace_id": "1234",
            "basespace_name": "test\tproject",
            "basespace_description": "",
            "basespace_date_created": (
                datetime.utcnow() - timedelta(days=7)
            ).isoformat(),
        }
    ],
)


def test_basespace_list__no_permission(mocker):
    """Test BaseSpace projects no permission available to show them."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_list_basespace_projects = mocker.patch.object(
        APIClient,
        "list_basespace_projects",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=403
        ),
        return_value={"detail": "Not found"},
    )
    res = runner.invoke(basespace_list, ["--email", "foo@bar.com", "--password", "123"])
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_list_basespace_projects.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\n".join(
            [
                "ERROR: There was an error listing BaseSpace projects.",
                "ERROR: You do not have the sufficient permission "
                "level required to perform this operation.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


def test_basespace_list__slow_response_retry_list(mocker):
    """Test BaseSpace projects slow response retry on the list."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_list_basespace_projects = mocker.patch.object(
        APIClient,
        "list_basespace_projects",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )
    res = runner.invoke(basespace_list, ["--email", "foo@bar.com", "--password", "123"])
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    assert mocked_list_basespace_projects.call_count == 2


def test_basespace_list(mocker):
    """Test BaseSpace projects being outputed to the shell."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_list_basespace_projects = mocker.patch.object(
        APIClient,
        "list_basespace_projects",
        return_value=BaseSpaceProject(**MOCKED_BASESPACE_PROJECTS),
    )
    res = runner.invoke(basespace_list, ["--email", "foo@bar.com", "--password", "123"])
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_list_basespace_projects.assert_called_once()

    basespace_project = BaseSpaceProjectDetail(
        **MOCKED_BASESPACE_PROJECTS["results"][0]
    )

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                str(basespace_project.basespace_date_created),
                str(basespace_project.basespace_id),
                basespace_project.basespace_name.replace("\t", " "),
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


def test_basespace_list__with_unexpected_keys(mocker):
    """Test BaseSpace projects being outputed to the shell with
    an unexpected key as a part of the response.
    """

    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_list_basespace_projects = mocker.patch.object(
        APIClient,
        "list_basespace_projects",
        return_value=BaseSpaceProject(**MOCKED_BASESPACE_PROJECTS_WITH_UNEXPECTED_KEYS),
    )
    res = runner.invoke(basespace_list, ["--email", "foo@bar.com", "--password", "123"])
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_list_basespace_projects.assert_called_once()

    basespace_project = BaseSpaceProjectDetail(
        **MOCKED_BASESPACE_PROJECTS_WITH_UNEXPECTED_KEYS["results"][0]
    )

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\t".join(
            [
                str(basespace_project.basespace_date_created),
                str(basespace_project.basespace_id),
                basespace_project.basespace_name.replace("\t", " "),
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()



"""Test BaseSpace's projects import command."""

from uuid import uuid4

from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)
from gencove.command.basespace.projects.cli import basespace_import


def test_basespace_import__bad_project_id(mocker):
    """Test BaseSpace import when non-uuid string is used as project
    id."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_import_basespace_projects = mocker.patch.object(
        APIClient,
        "import_basespace_projects",
    )
    res = runner.invoke(
        basespace_import,
        [
            "1111111",
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_not_called()
    mocked_import_basespace_projects.assert_not_called()
    assert "Project ID is not valid" in res.output


def test_basespace_import__bad_json(mocker):
    """Test BaseSpace import when bad JSON is used."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_import_basespace_projects = mocker.patch.object(
        APIClient,
        "import_basespace_projects",
    )
    res = runner.invoke(
        basespace_import,
        [
            "1111111",
            str(uuid4()),
            "--metadata-json",
            "{bad:",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_import_basespace_projects.assert_not_called()
    assert "Metadata JSON is not valid" in res.output


# the following test will work for any 400 case:
# (too many projects, no projects (cannot happen), parse error
# no samples in the projects or too many samples)
def test_basespace_import__no_pipeline_capabilities(mocker):
    """Test BaseSpace import failure when project has no
    pipeline capabilities.
    """
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_import_basespace_projects = mocker.patch.object(
        APIClient,
        "import_basespace_projects",
        side_effect=APIClientError(
            message="Project configuration must be set before adding "
            "samples to a project.",
            status_code=400,
        ),
    )

    res = runner.invoke(
        basespace_import,
        [
            "1111111",
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_import_basespace_projects.assert_called_once()
    assert "There was an error importing Biosamples from BaseSpace." in res.output
    assert (
        "Project configuration must be set before adding samples to a project."
        in res.output
    )


def test_basespace_import__not_owned_project(mocker):
    """Test BaseSpace import failure when project is not owned."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_import_basespace_projects = mocker.patch.object(
        APIClient,
        "import_basespace_projects",
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        basespace_import,
        [
            "1111111",
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_import_basespace_projects.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


def test_basespace_import__success_with_json(mocker):
    """Test BaseSpace import success with optional json passed."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_import_basespace_projects = mocker.patch.object(
        APIClient,
        "import_basespace_projects",
        return_value=None,
    )

    res = runner.invoke(
        basespace_import,
        [
            "1111111",
            str(uuid4()),
            "--metadata-json",
            '{"somekey": "somevalue"}',
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_import_basespace_projects.assert_called_once()
    assert "Request to import BaseSpace projects accepted." in res.output


def test_basespace_import__success_with_multiple(mocker):
    """Test BaseSpace import success with multiple BaseSpace projects."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_import_basespace_projects = mocker.patch.object(
        APIClient,
        "import_basespace_projects",
        return_value=None,
    )

    res = runner.invoke(
        basespace_import,
        [
            "1111111,2222222",
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_import_basespace_projects.assert_called_once()
    assert "Request to import BaseSpace projects accepted." in res.output


def test_basespace_import__success(mocker):
    """Test BaseSpace import success."""
    runner = CliRunner()
    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_import_basespace_projects = mocker.patch.object(
        APIClient,
        "import_basespace_projects",
        return_value=None,
    )

    res = runner.invoke(
        basespace_import,
        [
            "1111111",
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_import_basespace_projects.assert_called_once()
    assert "Request to import BaseSpace projects accepted." in res.output



"""Tests for command.explorer.data.common"""
import os
import sys
import uuid
from unittest import TestCase
from unittest.mock import MagicMock, patch

# pylint: disable=wrong-import-order, import-error

import pytest

from gencove.command.explorer.data.common import (  # noqa: I100
    ExplorerDataCredentials,
    GencoveExplorerManager,
    request_is_from_explorer,
)
from gencove.tests.utils import MOCK_UUID


class TestGencoveExplorerManager:  # pylint: disable=too-many-public-methods
    """Test GencoveExplorerManager"""

    @pytest.fixture(autouse=True)
    def setup(self):
        """Initialize mocked instance of GencoveExplorerManager"""
        # pylint: disable=attribute-defined-outside-init
        self.user_id = "11111111-1111-1111-1111-111111111111"
        self.organization_id = "11111111-1111-1111-1111-111111111111"
        self.org_id_short = self.organization_id.replace("-", "")[:12]
        self.mock_aws_credentials = ExplorerDataCredentials(
            access_key="mock_access",
            secret_key="mock_secret",
            token="mock_token",
            region_name="mock_region",
        )
        self.explorer_manager = GencoveExplorerManager(
            aws_session_credentials=self.mock_aws_credentials,
            user_id=self.user_id,
            organization_id=self.organization_id,
        )

    def test_bucket_name(self):
        """Test bucket_name"""
        assert (
            self.explorer_manager.bucket_name
            == f"gencove-explorer-{self.org_id_short}"  # noqa: E501 pylint: disable=line-too-long
        )

    def test_users_prefix(self):
        """Test users_prefix"""
        assert self.explorer_manager.users_prefix == "users"

    def test_base_prefix(self):
        """Test base_prefix"""
        expected_prefix = f"users/{self.user_id}"
        assert self.explorer_manager.base_prefix == expected_prefix

    def test_shared_prefix(self):
        """Test shared_prefix"""
        assert self.explorer_manager.shared_prefix == "shared"

    def test_data_org_prefix(self):
        """Test data_org_prefix"""
        expected_prefix = "shared/files"
        assert self.explorer_manager.data_org_prefix == expected_prefix

    def test_data_prefix(self):
        """Test data_prefix"""
        expected_prefix = f"users/{self.user_id}/files"
        assert self.explorer_manager.data_prefix == expected_prefix

    def test_user_prefix(self):
        """Test user_prefix"""
        expected_prefix = f"users/{self.user_id}/files"
        assert self.explorer_manager.user_prefix == expected_prefix

    def test_tmp_prefix(self):
        """Test tmp_prefix"""
        expected_prefix = f"users/{self.user_id}/files/tmp"
        assert self.explorer_manager.tmp_prefix == expected_prefix

    def test_tmp_org_prefix(self):
        """Test tmp_org_prefix"""
        expected_prefix = "shared/files/tmp"
        assert self.explorer_manager.tmp_org_prefix == expected_prefix

    def test_scratch_prefix(self):
        """Test scratch_prefix"""
        expected_prefix = f"users/{self.user_id}/scratch"
        assert self.explorer_manager.scratch_prefix == expected_prefix

    def test_user_scratch_s3_prefix(self):
        """Test user_scratch_s3_prefix"""
        expected_prefix = (
            f"s3://gencove-explorer-{self.org_id_short}/users/{self.user_id}/scratch"
        )
        assert self.explorer_manager.user_scratch_s3_prefix == expected_prefix

    def test_shared_org_s3_prefix(self):
        """Test shared_org_s3_prefix"""
        expected_prefix = f"s3://gencove-explorer-{self.org_id_short}/shared"
        assert self.explorer_manager.shared_org_s3_prefix == expected_prefix

    def test_data_gencove_s3_prefix(self):
        """Test data_gencove_s3_prefix"""
        expected_prefix = "s3://gencove-explorer-data/files"
        assert self.explorer_manager.data_gencove_s3_prefix == expected_prefix

    def test_data_org_s3_prefix(self):
        """Test data_org_s3_prefix"""
        expected_prefix = f"s3://gencove-explorer-{self.org_id_short}/shared/files"
        assert self.explorer_manager.data_org_s3_prefix == expected_prefix

    def test_data_s3_prefix(self):
        """Test data_s3_prefix"""
        expected_prefix = (
            f"s3://gencove-explorer-{self.org_id_short}/users/{self.user_id}/files"
        )
        assert self.explorer_manager.data_s3_prefix == expected_prefix

    def test_users_s3_prefix(self):
        """Test users_s3_prefix"""
        expected_prefix = f"s3://gencove-explorer-{self.org_id_short}/users"
        assert self.explorer_manager.users_s3_prefix == expected_prefix

    def test_aws_env(self):
        """Test aws_env"""
        expected = {
            "AWS_ACCESS_KEY_ID": "mock_access",
            "AWS_SECRET_ACCESS_KEY": "mock_secret",
            "AWS_SESSION_TOKEN": "mock_token",
            "AWS_DEFAULT_REGION": "mock_region",
            "AWS_REGION": "mock_region",
        }
        assert self.explorer_manager.aws_env == expected

    def test_empty_aws_env(self):
        """Test that aws_env returns an empty dict when no aws_session_credentials."""
        explorer_manager = GencoveExplorerManager(
            aws_session_credentials=None,
            user_id=self.user_id,
            organization_id=self.organization_id,
        )
        # pylint: disable=C1803 use-implicit-booleaness-not-comparison
        assert explorer_manager.aws_env == {}

    @pytest.mark.parametrize(
        "path, expected",
        [
            ("e://path", True),
            (None, False),
            ("invalid://path", False),
            ("s3://bucket/path", False),
        ],
    )
    def test_uri_ok(self, path, expected):
        """Test uri_ok()"""
        assert self.explorer_manager.uri_ok(path) == expected

    def test_translate_path_to_s3_path_valid(self):
        """Test translate_path_to_s3_path"""
        translated_path = self.explorer_manager.translate_path_to_s3_path(
            path="e://users/me"
        )
        assert (
            translated_path
            == f"s3://gencove-explorer-{self.org_id_short}/users/{self.organization_id}/files/"  # noqa: E501 pylint: disable=line-too-long
        )  # noqa: E501

    def test_translate_path_to_s3_path_invalid(self):
        """Test with an invalid e:// path and expect a ValueError"""
        with pytest.raises(ValueError):
            self.explorer_manager.translate_path_to_s3_path("e://invalid/path")

    @patch("gencove.command.explorer.data.common.GencoveExplorerManager.list_users")
    def test_list_users(self, mock_list_users):
        """Test list_users()"""
        mock_list_users.return_value = "user1\nuser2"
        self.explorer_manager.list_users()
        mock_list_users.assert_called_once()

    @patch("gencove.command.explorer.data.common.subprocess", create=True)
    def test_execute_aws_s3_src_dst(self, mocked_aws):  # pylint: disable= W0613
        """Test execute_aws_s3_src_dst()"""
        cmd = "cp"
        source = "e://users/me/source"
        destination = "e://users/me/destination"
        args = ["--arg1", "--arg2"]

        translated_source = self.explorer_manager.translate_path_to_s3_path(source)
        translated_destination = self.explorer_manager.translate_path_to_s3_path(
            destination
        )

        s3_cmd = self.explorer_manager.execute_aws_s3_src_dst(
            cmd, source, destination, args
        )

        assert s3_cmd == [cmd, translated_source, translated_destination, *args]
        user_s3_prefix = (
            f"s3://gencove-explorer-{self.org_id_short}/users/{self.user_id}/files"
        )
        env = self.explorer_manager.aws_env.copy()
        env["PATH"] = os.environ["PATH"]
        mocked_aws.run.assert_called_once_with(
            [
                "aws",
                "s3",
                cmd,
                f"{user_s3_prefix}/source",
                f"{user_s3_prefix}/destination",
                *args,
            ],
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
            env=env,
            check=True,
        )

    # ,

    @patch("gencove.command.explorer.data.common.subprocess", create=True)
    def test_execute_aws_s3_src_dst_invalid_path(
        self, mocked_aws
    ):  # noqa: E501 pylint: disable=W0613
        """Test execute_aws_s3_src_dst() with invalid src and dst"""
        cmd = "cp"
        source = "invalid"
        destination = "invalid"
        args = ["--arg1", "--arg2"]

        with pytest.raises(ValueError):
            self.explorer_manager.execute_aws_s3_src_dst(cmd, source, destination, args)

    @patch("gencove.command.explorer.data.common.subprocess", create=True)
    def test_execute_execute_aws_s3_path(self, mocked_aws):  # pylint: disable=W0613
        """Test execute_aws_s3_path()"""
        cmd = "cp"
        path = "e://users/me/source"
        args = ["--arg1", "--arg2"]

        translated_path = self.explorer_manager.translate_path_to_s3_path(path)

        s3_cmd = self.explorer_manager.execute_aws_s3_path(cmd, path, args)

        assert s3_cmd == [cmd, translated_path, *args]
        user_s3_prefix = (
            f"s3://gencove-explorer-{self.org_id_short}/users/{self.user_id}/files"
        )
        env = self.explorer_manager.aws_env.copy()
        env["PATH"] = os.environ["PATH"]
        mocked_aws.run.assert_called_once_with(
            ["aws", "s3", cmd, f"{user_s3_prefix}/source", *args],
            stdin=sys.stdin,
            stdout=sys.stdout,
            stderr=sys.stderr,
            env=env,
            check=True,
        )

    @patch("gencove.command.explorer.data.common.subprocess", create=True)
    def test_execute_execute_aws_s3_path_invalid_path(
        self, mocked_aws
    ):  # noqa: E501 pylint: disable=W0613
        """Test execute_aws_s3_path() with invalid path"""
        cmd = "cp"
        path = "invalid"
        args = ["--arg1", "--arg2"]

        with pytest.raises(ValueError):
            self.explorer_manager.execute_aws_s3_path(cmd, path, args)


class TestRequestIsFromExplorerInstance(TestCase):
    """Test case for request_is_from_instance()"""

    def setUp(self):
        self.user_id = MOCK_UUID
        self.account_id = "123412341234"

    def test_from_explorer_instance(self):
        """Test when request issued from explorer instance"""
        with patch(
            "gencove.command.explorer.data.common.boto3.client"
        ) as mock_boto3, patch.dict(os.environ, {"GENCOVE_USER_ID": self.user_id}):
            client_mock = MagicMock()
            client_mock.get_caller_identity.return_value = {
                "Arn": f"arn:aws:iam::{self.account_id}:role/explorer-user-{self.user_id}-role"  # noqa: E501 pylint: disable=line-too-long
            }
            mock_boto3.return_value = client_mock

            assert request_is_from_explorer() is True

    def test_from_explorer_cluster(self):
        """Test when request issued from explorer cluster"""
        with patch(
            "gencove.command.explorer.data.common.boto3.client"
        ) as mock_boto3, patch.dict(os.environ, {"GENCOVE_USER_ID": self.user_id}):
            user_id_dashes = str(uuid.UUID(self.user_id))
            client_mock = MagicMock()
            client_mock.get_caller_identity.return_value = {
                "Arn": f"arn:aws:iam::{self.account_id}:role/explorer-{user_id_dashes}-ecs_task_role"  # noqa: E501 pylint: disable=line-too-long
            }
            mock_boto3.return_value = client_mock

            assert request_is_from_explorer() is True

    def test_not_from_explorer(self):
        """Test when request not from explorer instance"""
        with patch(
            "gencove.command.explorer.data.common.boto3.client"
        ) as mock_boto3, patch.dict(os.environ, {"GENCOVE_USER_ID": self.user_id}):
            client_mock = MagicMock()
            client_mock.get_caller_identity.return_value = {
                "Arn": "arn:aws:iam::1111111111111111:role/some-other-role"
            }
            mock_boto3.return_value = client_mock

            assert request_is_from_explorer() is False

    def test_exception(self):
        """Test when exception is raised"""
        with patch(
            "gencove.command.explorer.data.common.boto3.client"
        ) as mock_boto3, patch.dict(os.environ, {"GENCOVE_USER_ID": self.user_id}):
            mock_boto3.side_effect = Exception("AWS error")

            assert request_is_from_explorer() is False



"""Test samples get metadata command."""

import io
import json
import sys
from uuid import uuid4

from click import echo
from click.testing import CliRunner

from gencove.client import APIClient, APIClientError, APIClientTimeout
from gencove.command.samples.cli import get_metadata
from gencove.models import SampleMetadata


def test_get_metadata__bad_sample_id(mocker):
    """Test get metadata failure when non-uuid string is used as
    sample id.
    """
    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_metadata = mocker.patch.object(
        APIClient,
        "get_metadata",
    )

    res = runner.invoke(
        get_metadata,
        [
            "1111111",
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )

    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_metadata.assert_not_called()
    assert "Sample ID is not valid" in res.output


def test_get_metadata__not_owned_sample(mocker):
    """Test get metadata failure when sample is not owned.
    This will return empty response to handle cases when using the CLI tool.
    """
    mocked_response = {"detail": "Not found."}

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_metadata = mocker.patch.object(
        APIClient,
        "get_metadata",
        return_value=mocked_response,
        side_effect=APIClientError(message="", status_code=403),
    )

    res = runner.invoke(
        get_metadata,
        [
            str(uuid4()),
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    mocked_get_metadata.assert_called_once()
    assert "You do not have the sufficient permission" in res.output


def test_get_metadata__empty(mocker):
    """Test sample doesn't have metadata."""
    sample_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    mocked_get_metadata = mocker.patch.object(
        APIClient,
        "get_metadata",
        return_value=SampleMetadata(
            metadata=None,
        ),
    )

    res = runner.invoke(
        get_metadata,
        [
            sample_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_metadata.assert_called_once()
    assert json.dumps(None, indent=4) in res.output


def test_get_metadata__success_custom_filename(mocker):
    """Test sample get metadata success to custom file."""
    sample_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    metadata = {"somekey": "somevalue"}
    mocked_response = {"metadata": metadata}
    mocked_get_metadata = mocker.patch.object(
        APIClient,
        "get_metadata",
        return_value=SampleMetadata(**mocked_response),
    )
    custom_filename = "result.json"
    with runner.isolated_filesystem():
        res = runner.invoke(
            get_metadata,
            [
                sample_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--output-filename",
                custom_filename,
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_get_metadata.assert_called_once()
        with open(custom_filename, "r", encoding="utf-8") as output_file:
            output_content = output_file.read()
        assert json.dumps(metadata, indent=4) == output_content


def test_get_metadata__success_nested_custom_filename(mocker):
    """Test sample get metadata success to custom file."""
    sample_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    metadata = {"somekey": "somevalue"}
    mocked_response = {"metadata": metadata}
    mocked_get_metadata = mocker.patch.object(
        APIClient,
        "get_metadata",
        return_value=SampleMetadata(**mocked_response),
    )
    custom_filename = "somefolder/result.json"
    with runner.isolated_filesystem():
        res = runner.invoke(
            get_metadata,
            [
                sample_id,
                "--email",
                "foo@bar.com",
                "--password",
                "123",
                "--output-filename",
                custom_filename,
            ],
        )
        assert res.exit_code == 0
        mocked_login.assert_called_once()
        mocked_get_metadata.assert_called_once()
        with open(custom_filename, "r", encoding="utf-8") as output_file:
            output_content = output_file.read()
        assert json.dumps(metadata, indent=4) == output_content


def test_get_metadata__success_stdout(mocker):
    """Test sample get metadata success to stdout."""
    sample_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    metadata = {"somekey": "somevalue"}
    mocked_response = {"metadata": metadata}
    mocked_get_metadata = mocker.patch.object(
        APIClient,
        "get_metadata",
        return_value=SampleMetadata(**mocked_response),
    )

    res = runner.invoke(
        get_metadata,
        [
            sample_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--output-filename",
            "-",
        ],
    )
    assert res.exit_code == 0
    mocked_login.assert_called_once()
    mocked_get_metadata.assert_called_once()
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(json.dumps(metadata, indent=4))
    assert output_line.getvalue() in res.output.encode()


def test_get_metadata__slow_response_retry(mocker):
    """Test sample get metadata slow response retry."""
    sample_id = str(uuid4())

    runner = CliRunner()

    mocked_login = mocker.patch.object(APIClient, "login", return_value=None)
    metadata = {"somekey": "somevalue"}
    mocked_get_metadata = mocker.patch.object(
        APIClient,
        "get_metadata",
        side_effect=APIClientTimeout("Could not connect to the api server"),
    )

    res = runner.invoke(
        get_metadata,
        [
            sample_id,
            "--email",
            "foo@bar.com",
            "--password",
            "123",
            "--output-filename",
            "-",
        ],
    )
    assert res.exit_code == 1
    mocked_login.assert_called_once()
    assert mocked_get_metadata.call_count == 2
    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(json.dumps(metadata, indent=4))
    assert output_line.getvalue() not in res.output.encode()



"""List uploads subcommand."""
import backoff

# pylint: disable=wrong-import-order
from gencove.client import APIClientError, APIClientTimeout  # noqa: I100
from gencove.command.base import Command

from .utils import get_line


class ListSampleSheet(Command):
    """List sample sheet command executor."""

    def __init__(self, credentials, options):
        super().__init__(credentials, options)
        self.status = options.status
        self.gncv_path = options.search

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""
        pass  # pylint: disable=unnecessary-pass

    def execute(self):
        self.echo_debug(
            "Retrieving sample sheet: "
            f"status={self.status} search_term={self.gncv_path}"
        )
        try:
            for uploads in self.get_paginated_sample_sheet():
                if not uploads:
                    self.echo_debug("No matching uploads found.")
                    return

                for upload in uploads:
                    self.echo_data(get_line(upload))
        except APIClientError as err:
            if err.status_code == 404:
                self.echo_error("Uploads do not exist.")
            raise

    def get_paginated_sample_sheet(self):
        """Paginate over all sample sheet pages.

        Yields:
            paginated lists of uploads
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get sample sheet page")
            resp = self.get_sample_sheet(next_link)
            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    def get_sample_sheet(self, next_link=None):
        """Get sample sheet page."""
        return self.api_client.get_sample_sheet(
            gncv_path=self.gncv_path,
            assigned_status=self.status,
            next_link=next_link,
        )



"""Monthly usage shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import MonthlyUsageReport


@click.command("monthly-usage")
@click.option(
    "--from",
    "from_",
    help="Date to start report from in format YYYY-MM, e.g. 2023-01. "
    "If supplied, --to parameter must also be supplied.",
    type=str,
    default=None,
    required=False,
)
@click.option(
    "--to",
    help="Date to end report at in format YYYY-MM, e.g. 2023-03. "
    "If supplied, --from parameter must also be supplied.",
    type=str,
    default=None,
    required=False,
)
@click.option(
    "--output-filename",
    help="Output filename for the output report. "
    "If not supplied, will download the file to the current directory.",
    type=click.Path(),
    default=None,
    required=False,
)
@add_options(common_options)
def monthly_usage(  # pylint: disable=too-many-arguments,invalid-name
    from_,
    to,
    output_filename,
    host,
    email,
    password,
    api_key,
):
    """Get monthly usage report for organization.
    If --to and --from parameters are not supplied,
    the last 12 months of usage are retrieved."""
    MonthlyUsageReport(
        from_,
        to,
        output_filename,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Get Reference Genome command executor."""
import re

import backoff

from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.download.utils import (
    _create_filepath,
    download_file,
    get_filename_from_download_url,
)
from gencove.command.utils import validate_file_types
from gencove.constants import FileTypesObject
from gencove.exceptions import ValidationError


class GetReferenceGenome(Command):
    """Get Reference Genome command executor."""

    def __init__(
        self, project_id, destination, file_types, credentials, options, no_progress
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.destination = destination
        self.file_types = file_types
        self.no_progress = no_progress

    def initialize(self):
        """Initialize subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""
        try:
            valid_file_types = self.api_client.get_file_types(
                project_id=self.project_id,
                object_param=FileTypesObject.REFERENCE_GENOME.value,
            ).results
            invalid_file_types = validate_file_types(self.file_types, valid_file_types)
            if invalid_file_types:
                raise ValidationError(
                    f"Invalid file types: {', '.join(invalid_file_types)}. "
                    "Run 'gencove file-types --object reference-genome' command for a "
                    "list of valid file types. "
                    "Use with --project-id option to see project file types."
                )
        except client.APIClientError as err:
            self.echo_debug(err)
            self.echo_warning("There was an error while validating file types.")

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=5,
        max_time=30,
    )
    @backoff.on_exception(
        backoff.expo,
        client.APIClientTooManyRequestsError,
        max_tries=20,
    )
    def execute(self):
        """Download Reference Genome files for a given project."""
        self.echo_debug(
            f"Downloading Reference Genome files for project {self.project_id}"
        )
        try:
            project = self.api_client.get_project(self.project_id)
            self.echo_debug(project)
            file_types_re = re.compile("|".join(self.file_types), re.IGNORECASE)
            for file in project.files:
                if self.file_types and not file_types_re.match(file.file_type):
                    self.echo_debug(
                        "Deliverable file type is not in desired file types"
                    )
                    continue
                self.echo_debug(f"Downloading {file.file_type}, id {file.id}")
                filename = get_filename_from_download_url(file.download_url)
                file_path = _create_filepath(self.destination, "", filename)
                self.echo_debug(f"file path is: {file_path}")
                download_file(
                    file_path,
                    file.download_url,
                    no_progress=self.no_progress,
                )
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 404:
                self.echo_error(f"Project {self.project_id} does not exist.")
            raise



"""Commands to be executed from command line."""
import click

from .archive.cli import archive
from .cp.cli import cp
from .ls.cli import ls
from .presign.cli import presign
from .restore.cli import restore
from .rm.cli import rm
from .sync.cli import sync


@click.group()
def data():
    """Explorer data management commands."""


data.add_command(archive)
data.add_command(ls)
data.add_command(cp)
data.add_command(restore)
data.add_command(rm)
data.add_command(sync)
data.add_command(presign)

if __name__ == "__main__":
    data()



"""Utilities for processing and validating uploads from sample sheet."""


def get_line(upload):
    """Build a list of relevant data to be printed.

    Args:
        upload (dict): object from sample sheet

    Returns:
        list(str): list of relevant data to be printed
    """
    parts = [
        str(upload.client_id),
        str(upload.fastq.r1.upload),
        upload.fastq.r1.destination_path,
    ]
    if upload.fastq.r2:
        parts.append(str(upload.fastq.r2.upload))
        parts.append(upload.fastq.r2.destination_path)

    return "\t".join(parts)



"""Configure explorer data presign subcommand."""
import sys

import click

from ..common import (
    GencoveExplorerManager,
    request_is_from_explorer,
    validate_explorer_user_data,
)
from ....base import Command
from ....utils import user_has_aws_in_path


class Presign(Command):
    """Presign Explorer object command executor."""

    def __init__(self, ctx, path, credentials, options):
        super().__init__(credentials, options)
        self.ctx = ctx
        self.path = path

        # Populated after self.login() is called
        self.user = None
        self.organization = None
        self.aws_session_credentials = None

    def validate(self):
        """Validate presign"""
        validate_explorer_user_data(self.user, self.organization)
        user_has_aws_in_path(raise_exception=True)

    def initialize(self):
        """Initialize presign subcommand."""
        self.login()
        self.user = self.api_client.get_user_details()
        self.organization = self.api_client.get_organization_details()

        if not request_is_from_explorer():
            self.aws_session_credentials = (
                self.api_client.get_explorer_data_credentials()
            )

    def execute(self):
        """Make a request to presign Explorer object."""
        self.echo_debug("Presign Explorer object.")

        explorer_manager = GencoveExplorerManager(
            aws_session_credentials=self.aws_session_credentials,
            user_id=str(self.user.id),
            organization_id=str(self.organization.id),
        )

        if self.path == explorer_manager.EXPLORER_SCHEME:
            click.echo(
                "Please provide a valid path to an Explorer object",
                err=True,
            )
            sys.exit(1)
        if (
            self.path.rstrip("/")
            == f"{explorer_manager.EXPLORER_SCHEME}{explorer_manager.USERS}"
        ):
            click.echo(
                "Please provide a valid path to an Explorer object",
                err=True,
            )
            sys.exit(1)
        explorer_manager.execute_aws_s3_path("presign", self.path, self.ctx.args)



"""Python library which enables you to use Gencoves' research backend."""
import click

from gencove import version
from gencove.command.basespace import basespace
from gencove.command.download import download
from gencove.command.explorer import explorer
from gencove.command.files import list_file_types
from gencove.command.projects import projects
from gencove.command.reports import reports
from gencove.command.s3_imports import s3
from gencove.command.sample_manifests import sample_manifests
from gencove.command.samples import samples
from gencove.command.upload import upload
from gencove.command.uploads import uploads
from gencove.command.webhook import webhooks
from gencove.utils import python_version_check


def announcements():
    """Preamble announcements displayed whenever the CLI is called"""
    python_version_check()


@click.group()
@click.version_option(version=version.version())
def cli():
    """Gencove's command line interface."""


announcements()
cli.add_command(basespace)
cli.add_command(download)
cli.add_command(explorer)
cli.add_command(list_file_types)
cli.add_command(upload)
cli.add_command(uploads)
cli.add_command(projects)
cli.add_command(reports)
cli.add_command(samples)
cli.add_command(sample_manifests)
cli.add_command(s3)
cli.add_command(webhooks)

if __name__ == "__main__":
    cli()



"""Tests upload command of Gencove CLI."""
# pylint: disable=wrong-import-order, import-error

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError
from gencove.command.basespace.autoimports.create.cli import (
    create,
)
from gencove.tests.decorators import assert_authorization, assert_no_requests
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.utils import MOCK_UUID

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/basespace/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
        ],
        "before_record_response": [
            filter_jwt,
        ],
    }


@assert_no_requests
def test_basespace_autoimport_create_project_id_not_uuid(
    credentials,
    mocker,
):
    """Test that project id is valid UUID when creating an automated import."""
    runner = CliRunner()

    mocked_autoimport_from_basespace = mocker.patch.object(
        APIClient,
        "autoimport_from_basespace",
    )
    res = runner.invoke(
        create,
        [
            "1234",
            "identifier",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_autoimport_from_basespace.assert_not_called()
    assert "ERROR: Project ID is not valid. Exiting." in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_basespace_autoimport_create_invalid_metadata(credentials, mocker, project_id):
    """Test that passed optional metadata is valid when creating an
    automated import.
    """
    runner = CliRunner()

    mocked_autoimport_from_basespace = mocker.patch.object(
        APIClient,
        "autoimport_from_basespace",
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "identifier",
            "--metadata-json",
            "[1,2,3",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_autoimport_from_basespace.assert_not_called()
    assert "ERROR: Metadata JSON is not valid. Exiting." in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_basespace_autoimport_create_no_permission(credentials, mocker, project_id):
    """Test that user cannot create an automated import if no permissions."""
    runner = CliRunner()

    mocked_autoimport_from_basespace = mocker.patch.object(
        APIClient,
        "autoimport_from_basespace",
        side_effect=APIClientError(message="", status_code=403),
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "identifier",
            *credentials,
        ],
    )

    assert res.exit_code == 1
    mocked_autoimport_from_basespace.assert_called_once()
    assert "There was an error creating a periodic import job" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
@pytest.mark.parametrize("action", ["create", "update"])
def test_basespace_autoimport_update_with_empty_metadata(
    credentials,
    mocker,
    project_id,
    action,
):
    """Test that user can pass empty metadata when creating an
    automated import.
    """
    runner = CliRunner()

    mocked_autoimport_from_basespace = mocker.patch.object(
        APIClient,
        "autoimport_from_basespace",
        return_value={"id": MOCK_UUID, "metadata": None, "action": action},
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "identifier",
            "--metadata-json",
            None,
            *credentials,
        ],
    )

    assert res.output == (
        f"Request to {action} a periodic import job of BaseSpace "
        "projects accepted.\n"
    )
    assert res.exit_code == 0
    mocked_autoimport_from_basespace.assert_called_once()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
@pytest.mark.parametrize("action", ["create", "update"])
def test_basespace_autoimport_with_metadata(credentials, mocker, project_id, action):
    """Test that user can pass optional metadata when creating an
    automated import.
    """
    runner = CliRunner()

    mocked_autoimport_from_basespace = mocker.patch.object(
        APIClient,
        "autoimport_from_basespace",
        return_value={
            "id": MOCK_UUID,
            "metadata": {"some": "metadata"},
            "action": action,
        },
    )
    res = runner.invoke(
        create,
        [
            project_id,
            "identifier",
            "--metadata-json",
            "[1,2,3]",
            *credentials,
        ],
    )

    assert res.output == (
        "Metadata will be assigned to the imported Biosamples.\n"
        f"Request to {action} a periodic import job of BaseSpace "
        "projects accepted.\n"
    )
    assert res.exit_code == 0
    mocked_autoimport_from_basespace.assert_called_once()


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
@pytest.mark.parametrize("action", ["create", "update"])
def test_basespace_autoimport_update_without_metadata(
    credentials, mocker, project_id, action
):
    """Test that user can create an import job without passing metadata."""
    runner = CliRunner()

    mocked_autoimport_from_basespace = mocker.patch.object(
        APIClient,
        "autoimport_from_basespace",
        return_value={"id": MOCK_UUID, "action": action},
    )

    res = runner.invoke(
        create,
        [
            project_id,
            "identifier",
            *credentials,
        ],
    )

    assert res.output == (
        f"Request to {action} a periodic import job of BaseSpace "
        "projects accepted.\n"
    )
    assert res.exit_code == 0
    mocked_autoimport_from_basespace.assert_called_once()



"""Commands to be executed from command line."""
import click

from .inactivity_stop.cli import inactivity_stop
from .list.cli import list_instances
from .shell.cli import shell
from .start.cli import start
from .stop.cli import stop


@click.group()
def instances():
    """Explorer instances management commands."""


instances.add_command(start)
instances.add_command(stop)
instances.add_command(list_instances)
instances.add_command(inactivity_stop)
instances.add_command(shell)



"""Test instances list command."""
# pylint: disable=wrong-import-order, import-error
import io
import sys


from click import echo
from click.testing import CliRunner

from gencove.client import (
    APIClient,
    APIClientError,
)  # noqa: I100
from gencove.command.explorer.instances.cli import list_instances
from gencove.models import (
    ExplorerInstance,  # noqa: I101
    ExplorerInstances,
)
from gencove.tests.decorators import assert_authorization
from gencove.tests.explorer.vcr.filters import (  # noqa: I101
    filter_list_instances_response,
    filter_instance_ids_request,
)
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/explorer/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_instance_ids_request,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_list_instances_response,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_empty(mocker, credentials):
    """Test user organization has no instances."""
    runner = CliRunner()
    mocked_get_instances = mocker.patch.object(
        APIClient,
        "get_explorer_instances",
        return_value=ExplorerInstances(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(list_instances, credentials)
    assert res.exit_code == 0
    mocked_get_instances.assert_called_once()
    assert "" in res.output


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_list_instances_no_permission(mocker, credentials):
    """Test instances no permission available to show them."""
    runner = CliRunner()
    mocked_get_instances = mocker.patch.object(
        APIClient,
        "get_explorer_instances",
        side_effect=APIClientError(
            message="API Client Error: Not Found: Not found.", status_code=403
        ),
        return_value={"detail": "Not found"},
    )
    res = runner.invoke(list_instances, credentials)
    assert res.exit_code == 1
    mocked_get_instances.assert_called_once()

    output_line = io.BytesIO()
    sys.stdout = output_line
    echo(
        "\n".join(
            [
                "ERROR: You do not have the sufficient permission "
                "level required to perform this operation.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!",
            ]
        )
    )
    assert output_line.getvalue() == res.output.encode()


@pytest.mark.vcr
@assert_authorization
def test_list_instances(mocker, credentials, recording, vcr):
    """Test instances being outputed to the shell."""
    runner = CliRunner()
    if not recording:
        # Mock list_instances only if using the cassettes, since we mock the
        # return value.
        list_instances_response = get_vcr_response("/api/v2/explorer-instances/", vcr)
        mocked_get_instances = mocker.patch.object(
            APIClient,
            "get_explorer_instances",
            return_value=ExplorerInstances(**list_instances_response),
        )
    res = runner.invoke(list_instances, credentials)
    assert res.exit_code == 0
    if not recording:
        mocked_get_instances.assert_called_once()
        instances = list_instances_response["results"]
        output_line = io.BytesIO()
        sys.stdout = output_line
        for instance in instances:
            instance = ExplorerInstance(**instance)
            echo(
                "\t".join(
                    [
                        str(instance.id),
                        str(instance.status),
                    ]
                )
            )
        assert output_line.getvalue() == res.output.encode()



"""Test project's batches list command."""
# pylint: disable=wrong-import-order, import-error
import operator
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientTimeout  # noqa: I100
from gencove.command.download.utils import download_file
from gencove.command.projects.cli import get_batch
from gencove.models import BatchDetail
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import (
    filter_aws_headers,
    filter_jwt,
    mock_binary_response,
    replace_gencove_url_vcr,
    replace_s3_from_url,
)
from gencove.tests.projects.vcr.filters import (
    filter_batches_request,
    filter_batches_response,
)
from gencove.tests.upload.vcr.filters import filter_volatile_dates
from gencove.tests.utils import get_vcr_response

from pydantic.networks import HttpUrl

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/projects/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_batches_request,
            replace_s3_from_url,
        ],
        "before_record_response": [
            filter_jwt,
            filter_volatile_dates,
            filter_batches_response,
            mock_binary_response,
            filter_aws_headers,
        ],
    }


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_get_batch__empty(credentials, mocker):
    """Test project has not batches."""
    runner = CliRunner()
    batch_id = str(uuid4())
    mocked_get_batch = mocker.patch.object(
        APIClient,
        "get_batch",
        return_value=BatchDetail(
            id=batch_id,
            name="cli-test-1",
            batch_type="batch-type-1",
            sample_ids=[str(uuid4()), str(uuid4())],
            last_status=dict(
                id=str(uuid4()),
                status="running",
                created="2020-08-02T22:13:54.547167Z",
            ),
            files=[],
        ),
    )

    res = runner.invoke(get_batch, [batch_id, *credentials])
    assert res.exit_code == 1
    mocked_get_batch.assert_called_once()
    assert res.output == (
        "ERROR: There are no deliverables available for batch"
        f" {batch_id}.\nAborted!\n"
    )


@pytest.mark.vcr
@assert_authorization
def test_get_batch__not_empty(batch_id, credentials, mocker, recording, vcr):
    """Test project batches being outputed to the shell."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock get_batch only if using the cassettes, since we mock the
            # return value.
            get_batch_response = get_vcr_response(
                "/api/v2/batches/", vcr, operator.contains
            )
            mocked_get_batch = mocker.patch.object(
                APIClient,
                "get_batch",
                return_value=BatchDetail(**get_batch_response),
            )
            mocked_download_file = mocker.patch(
                "gencove.command.projects.get_batch.main.download.utils."
                "download_file",
                side_effect=download_file,
            )
        res = runner.invoke(
            get_batch,
            [
                batch_id,
                *credentials,
                "--output-filename",
                "test.zip",
            ],
        )
    assert res.exit_code == 0

    if not recording:
        mocked_get_batch.assert_called_once()
        mocked_download_file.assert_called_once_with(
            "test.zip",
            HttpUrl(
                url="https://example.com/report.zip",
                scheme="https",
                host="example.com",
            ),
            no_progress=False,
        )


@pytest.mark.vcr
@assert_authorization
def test_get_batch__no_progress_not_empty(
    batch_id, credentials, mocker, recording, vcr
):
    """Test project batches being outputed to the shell without progress."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        if not recording:
            # Mock get_batch only if using the cassettes, since we mock the
            # return value.
            get_batch_response = get_vcr_response(
                "/api/v2/batches/", vcr, operator.contains
            )
            mocked_get_batch = mocker.patch.object(
                APIClient,
                "get_batch",
                return_value=BatchDetail(**get_batch_response),
            )
            mocked_download_file = mocker.patch(
                "gencove.command.projects.get_batch.main.download.utils."
                "download_file",
                side_effect=download_file,
            )
        res = runner.invoke(
            get_batch,
            [
                batch_id,
                *credentials,
                "--output-filename",
                "test.zip",
                "--no-progress",
            ],
        )
    assert res.exit_code == 0
    if not recording:
        mocked_get_batch.assert_called_once()
        mocked_download_file.assert_called_once_with(
            "test.zip",
            HttpUrl(
                url="https://example.com/report.zip",
                scheme="https",
                host="example.com",
            ),
            no_progress=True,
        )


@pytest.mark.default_cassette("jwt-create.yaml")
@pytest.mark.vcr
@assert_authorization
def test_get_batch__not_empty__slow_response_retry(credentials, mocker):
    """Test project batches slow response retry."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        batch_id = str(uuid4())
        mocked_get_batch = mocker.patch.object(
            APIClient,
            "get_batch",
            side_effect=APIClientTimeout("Could not connect to the api server"),
        )
        mocked_download_file = mocker.patch(
            "gencove.command.projects.get_batch.main.download.utils.download_file"
        )
        res = runner.invoke(
            get_batch,
            [
                batch_id,
                *credentials,
                "--output-filename",
                "test.zip",
            ],
        )
    assert res.exit_code == 1
    assert mocked_get_batch.call_count == 2
    mocked_download_file.assert_not_called()



"""Configure inactivity stop for explorer instances command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import StopInstances


@click.command("stop")
@add_options(common_options)
def stop(  # pylint: disable=too-many-arguments
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Stop instances."""  # noqa: E501
    StopInstances(
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Commands to be executed from command line."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.projects.get_reference_genome.main import GetReferenceGenome
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, Optionals
from gencove.logger import echo_debug


@click.command("get-reference-genome")
@click.argument("project_id", callback=validate_uuid)
@click.argument(
    "destination",
    type=click.Path(file_okay=False),
)
@click.option(
    "--file-types",
    help=(
        "A comma separated list of deliverable file types to download. "
        "If not specified, all file types will be downloaded."
    ),
)
@add_options(common_options)
@click.option(
    "--no-progress",
    is_flag=True,
    help="If specified, no progress bar is shown.",
)
def get_reference_genome(
    project_id,
    destination,
    file_types,
    host,
    email,
    password,
    api_key,
    no_progress,
):
    """Download Reference genomes of a project

    `PROJECT_ID`: Gencove project ID
    `DESTINATION`: path/to/save/deliverables/to

    Examples:

        Download all reference genome files:

            gencove projects get_reference_genome d9eaa54b-aaac-4b85-92b0-0b564be6d7db
             ./genome


        Download only fasta reference genome file:

            gencove projects get_reference_genome d9eaa54b-aaac-4b85-92b0-0b564be6d7db
             . --file-types genome-fasta
    """
    f_types = tuple()
    if file_types:
        f_types = tuple(f_type.strip() for f_type in file_types.split(","))
        echo_debug(f"File types translation: {f_types}")
    GetReferenceGenome(
        project_id,
        destination,
        f_types,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
        no_progress,
    ).run()



"""Configure explorer data cp subcommand."""
from ..common import (
    GencoveExplorerManager,
    request_is_from_explorer,
    validate_explorer_user_data,
)
from ....base import Command
from ....utils import user_has_aws_in_path


class Copy(Command):
    """Copy Explorer contents command executor."""

    def __init__(self, ctx, source, destination, credentials, options):
        super().__init__(credentials, options)
        self.ctx = ctx
        self.source = source
        self.destination = destination

        # Populated after self.login() is called
        self.user = None
        self.organization = None
        self.aws_session_credentials = None

    def validate(self):
        """Validate cp"""
        validate_explorer_user_data(self.user, self.organization)
        user_has_aws_in_path(raise_exception=True)

    def initialize(self):
        """Initialize cp subcommand."""
        self.login()
        self.user = self.api_client.get_user_details()
        self.organization = self.api_client.get_organization_details()

        if not request_is_from_explorer():
            self.aws_session_credentials = (
                self.api_client.get_explorer_data_credentials()
            )

    def execute(self):
        """Make a request to cp."""
        self.echo_debug("Copy Explorer contents.")

        explorer_manager = GencoveExplorerManager(
            aws_session_credentials=self.aws_session_credentials,
            user_id=str(self.user.id),
            organization_id=str(self.organization.id),
        )
        explorer_manager.execute_aws_s3_src_dst(
            "cp", self.source, self.destination, self.ctx.args
        )



"""Project pipelines list shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import ListPipelines


@click.command("list-pipelines")
@add_options(common_options)
def list_project_pipelines(host, email, password, api_key):
    """List pipelines that are available when creating a project."""
    ListPipelines(
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Project get sample manifests shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_destination_exists, validate_uuid
from gencove.constants import Credentials, Optionals

from .main import GetSampleManifests


@click.command("get-sample-manifests")
@click.argument("project_id", callback=validate_uuid)
@click.argument("destination", callback=validate_destination_exists)
@add_options(common_options)
def get_sample_manifests(  # pylint: disable=too-many-arguments
    project_id,
    destination,
    host,
    email,
    password,
    api_key,
):
    """Download all sample manifests in a project to a destination directory.

    `PROJECT_ID`: Gencove project ID

    `DESTINATION`: path/to/save/sample/manifests/to
    """
    GetSampleManifests(
        project_id,
        destination,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""List project's pipelines subcommand."""

import backoff

from gencove import client
from gencove.command.base import Command

from .utils import get_line


class ListPipelines(Command):
    """List pipelines command executor."""

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""

    def execute(self):
        self.echo_debug("Retrieving possible pipelines:")

        try:
            for pipelines in self.get_paginated_pipelines():
                if not pipelines:
                    self.echo_debug("No matching pipelines were found.")
                    return

                for pipeline in pipelines:
                    self.echo_data(get_line(pipeline))
        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error listing pipelines.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            else:
                raise

    def get_paginated_pipelines(self):
        """Paginate over all pipelines for the destination.

        Yields:
            paginated lists of pipelines
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get pipelines page")
            resp = self.get_pipelines(next_link)
            yield resp.results
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_pipelines(self, next_link=None):
        """Get pipelines page."""
        return self.api_client.get_pipelines(next_link=next_link)



"""Project autoimport Biosamples from BaseSpace projects shell command
definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.constants import Credentials, Optionals

from .main import BaseSpaceAutoImportList


@click.command("list")
@add_options(common_options)
def autoimport_list(
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Lists periodic import of BaseSpace projects (their Biosamples) jobs.

    Examples:

        List automatic import jobs of BaseSpace projects:

            gencove basespace autoimports list
    """  # noqa: E501

    BaseSpaceAutoImportList(
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



"""Project autoimport from S3 shell command definition.
"""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials

from .constants import S3AutoImportOptionals
from .main import S3AutoImport


@click.command("create")
@click.argument("project_id", callback=validate_uuid)
@click.argument("s3_uri")
@click.option(
    "--metadata-json",
    required=False,
    default=None,
    help=("Add metadata to all samples that are to be imported from S3 to a project."),
)
@add_options(common_options)
def create(  # pylint: disable=too-many-arguments
    project_id,
    s3_uri,
    metadata_json,
    host,
    email,
    password,
    api_key,
):  # pylint: disable=line-too-long
    """Sets up automatic import from S3 URI to a project in Gencove. Optionally assign metadata to the samples to be added when the automatic
    import job runs.

    `PROJECT_ID`: Gencove project ID

    `S3_URI`: s3 prefix where samples are and will be located

    Examples:

        Set up automatic import from S3 to a project:

            gencove s3 autoimports create 06a5d04b-526a-4471-83ba-fb54e0941758 s3://bucket/path/to/project

        Set up automatic import from S3 to a project with metadata:

            gencove s3 autoimports create 06a5d04b-526a-4471-83ba-fb54e0941758 s3://bucket/path/to/project --metadata-json='{"batch": "batch1"}'
    """  # noqa: E501

    S3AutoImport(
        project_id,
        s3_uri,
        Credentials(email=email, password=password, api_key=api_key),
        S3AutoImportOptionals(host=host, metadata_json=metadata_json),
    ).run()



"""Utilities for processing BaseSpace Biosamples."""
from gencove.command.utils import sanitize_string


def get_line(biosample):
    """Build a list of relevant data to be printed.

    Args:
        biosample (BaseSpaceBiosample): instance of BaseSpaceBiosample

    Returns:
        list(str): list of relevant data to be printed
    """
    return "\t".join(
        [
            str(biosample.basespace_date_created),
            str(biosample.basespace_id),
            sanitize_string(biosample.basespace_bio_sample_name),
        ]
    )



"""Packaging for Gencove CLI."""
from setuptools import find_packages, setup


def version():
    """Return current package version."""
    with open("gencove/version/A-major", "rt") as f:
        major = f.read().replace("\n", "")
    with open("gencove/version/B-minor", "rt") as f:
        minor = f.read().replace("\n", "")
    with open("gencove/version/C-patch", "rt") as f:
        patch = f.read().replace("\n", "")

    return f"{major}.{minor}.{patch}"


def long_description():
    with open("gencove/description/pypi_readme.md") as f:
        text = f.read()
    return text


setup(
    name="gencove",
    description="Gencove API and CLI tool",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    url="https://docs.gencove.com",
    author="Tomaz Berisa",
    license="Apache 2.0",
    version=version(),
    python_requires=">=3.7",
    packages=find_packages(exclude=["tests"]),
    install_requires=[
        # For the python-dateutil requirement, see:
        # https://github.com/awslabs/aws-shell/issues/161
        # and
        # https://stackoverflow.com/questions/27630114/matplotlib-issue-on-os-x-importerror-cannot-import-name-thread
        "python-dateutil>=2.2.0",
        # This six requirement is related to the python-dateutil requirement
        # above - we are attempting to keep the requirement as loose as
        # possible in terms of minimal required version
        "six>=1.5",
        "Click>=7.0",
        "requests>=2.19.1",
        "boto3>=1.17.97",
        "progressbar2==3.55.0",
        "backoff<=2.2.1",
        "pydantic==1.10.13",
        "click-default-group>=1.2.4",
    ],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "pytest-mock"],
    entry_points="""
        [console_scripts]
        gencove=gencove.cli:cli
        d=gencove.command.explorer.data.cli:data
    """,
    package_data={"gencove": ["version/*", "description/*"]},
)



"""Import existing samples to a project subcommand."""
import json
from typing import Generator

import backoff

from .utils import get_line
from ...base import Command
from ...utils import is_valid_json, is_valid_uuid
from .... import client
from ....constants import IMPORT_BATCH_SIZE, SampleArchiveStatus, SampleStatus
from ....exceptions import ValidationError
from ....models import ProjectSamples, SampleDetails
from ....utils import batchify


class ImportExistingSamples(Command):
    """Import existing samples command executor."""

    def __init__(
        self, project_id, source_project_id, source_sample_ids, credentials, options
    ):
        super().__init__(credentials, options)
        self.project_id = project_id
        self.source_project_id = source_project_id
        self.source_sample_ids = source_sample_ids
        self.metadata_json = options.metadata_json

    def initialize(self):
        """Initialize import-existing-samples subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """
        if self.metadata_json and is_valid_json(self.metadata_json) is False:
            raise ValidationError("Metadata JSON is not valid. Exiting.")
        if self.source_project_id and not is_valid_uuid(self.source_project_id):
            raise ValidationError("Source project ID is not valid. Exiting.")
        if (self.source_sample_ids and self.source_project_id) or (
            not self.source_sample_ids and not self.source_project_id
        ):
            raise ValidationError(
                "Either --source-project-id or --sample-ids option must be"
                " provided, but not both."
            )
        if self.project_id == self.source_project_id:
            raise ValidationError("Source and destination project must be different.")

    def execute(self):
        """Import existing samples to the given project."""
        if self.source_project_id:
            self.echo_debug(
                "No samples given, importing all succeeded and available"
                f" samples from source project {self.source_project_id}."
            )
            self.source_sample_ids = []
            for sample in self.get_paginated_samples():
                self.source_sample_ids.append(str(sample.id))
            self.echo_debug(
                f"Samples to import from project: {len(self.source_sample_ids)}"
            )
        metadata = None
        if self.metadata_json is not None:
            metadata = json.loads(self.metadata_json)
            self.echo_info("Assigning metadata to the importing samples.")
        self.echo_info(f"Import existing samples to the project: {self.project_id}")
        try:
            # Import existing samples to the project optionally
            # passing metadata.
            for samples_batch in batchify(
                self.source_sample_ids, batch_size=IMPORT_BATCH_SIZE
            ):
                import_existing_samples_response = (
                    self.api_client.import_existing_samples(
                        self.project_id, samples_batch, metadata
                    )
                )
                for imported_sample in import_existing_samples_response.samples:
                    self.echo_data(get_line(imported_sample))
            self.echo_info(
                f"Number of samples imported into the project {self.project_id}: "
                f"{len(self.source_sample_ids)}"
            )
            if metadata:
                self.echo_info("Metadata attached to each sample.")
        except client.APIClientError as err:
            self.echo_debug(err.message)
            self.echo_error("There was an error importing samples.")
            raise

    def get_paginated_samples(
        self,
    ) -> Generator[SampleDetails, None, None]:
        """Paginate over all completed samples for the destination.

        Yields:
            Samples in succeeded or failed_qc state that have files.
        """
        more = True
        next_link = None
        while more:
            self.echo_debug("Get all completed samples")
            resp = self.get_samples(next_link)
            if resp.results:
                for sample in resp.results:
                    if sample.last_status.status in ["failed qc", "succeeded"]:
                        yield sample
            next_link = resp.meta.next
            more = next_link is not None

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_samples(self, next_link=None) -> ProjectSamples:
        """Get all completed samples page."""
        return self.api_client.get_project_samples(
            project_id=self.source_project_id,
            next_link=next_link,
            sample_status=SampleStatus.COMPLETED.value,
            sample_archive_status=SampleArchiveStatus.AVAILABLE.value,
        )



"""Get project batch's deliverable subcommand."""

import backoff

# pylint: disable=wrong-import-order
from gencove import client  # noqa: I100
from gencove.command.base import Command
from gencove.command.utils import is_valid_uuid
from gencove.exceptions import ValidationError

from ... import download


class GetBatch(Command):
    """Get batch command executor."""

    # pylint: disable=too-many-arguments
    def __init__(self, batch_id, output_filename, credentials, options, no_progress):
        super().__init__(credentials, options)
        self.batch_id = batch_id
        self.output_filename = output_filename
        self.no_progress = no_progress

    def initialize(self):
        """Initialize list subcommand."""
        self.login()

    def validate(self):
        """Validate command input.

        Raises:
            ValidationError - if something is wrong with command parameters.
        """

        if is_valid_uuid(self.batch_id) is False:
            raise ValidationError("Batch ID is not valid. Exiting.")

    def execute(self):
        self.echo_debug("Retrieving batch:")

        try:
            batch = self.get_batch()
            if not batch.files:
                raise ValidationError(
                    f"There are no deliverables available for batch {self.batch_id}."
                )
            if len(batch.files) > 1:
                self.echo_warning(
                    "There is more than one deliverable available for "
                    f"batch {self.batch_id}."
                )
            deliverable = batch.files[0]
            download_path = (
                self.output_filename
                if self.output_filename
                else download.utils.get_filename_from_download_url(
                    deliverable.download_url
                )
            )
            download.utils.download_file(
                download_path,
                deliverable.download_url,
                no_progress=self.no_progress,
            )

        except client.APIClientError as err:
            self.echo_debug(err)
            if err.status_code == 400:
                self.echo_warning("There was an error getting the batch.")
                self.echo_info("The following error was returned:")
                self.echo_info(err.message)
            elif err.status_code == 404:
                self.echo_warning(
                    f"Batch {self.batch_id} does not exist.",
                )
            else:
                raise

    @backoff.on_exception(
        backoff.expo,
        (client.APIClientTimeout),
        max_tries=2,
        max_time=30,
    )
    def get_batch(self):
        """Get batches page."""
        return self.api_client.get_batch(batch_id=self.batch_id)



"""Utilities for processing Explorer instances."""

from typing import Union


def hours_to_human_readable(hours: int, from_org: Union[int, None] = None) -> str:
    """Turn hours into human readable format."""
    description = "disabled" if hours == 0 else ""
    if from_org:
        if description:
            description += " applied from organization"
        else:
            description = "applied from organization"
    description = f" ({description})" if description else ""
    return f"{hours}{description}"


def calculate_applied_hours_to_instance(instance_config, org_config):
    """Calculate applied hours to instance.

    Args:
        instance_config (dict): Instance config.
        org_config (dict): Org config.

    Returns:
        tuple(int, bool): Hours and boolean indicating if config was applied from
            org or not.
    """
    if (
        instance_config["stop_after_inactivity_hours"] is not None
        and not org_config["explorer_override_stop_after_inactivity_hours"]
    ):
        hours, from_org = instance_config["stop_after_inactivity_hours"], False
    else:
        hours, from_org = org_config["explorer_stop_after_inactivity_hours"], True
    return hours, from_org



"""Test file types list command"""
# pylint: disable=wrong-import-order, import-error
import operator
from uuid import uuid4

from click.testing import CliRunner

from gencove.client import APIClient, APIClientError
from gencove.command.files.cli import list_file_types
from gencove.models import FileTypesModel
from gencove.tests.decorators import assert_authorization
from gencove.tests.filters import filter_file_types_request
from gencove.tests.filters import filter_jwt, replace_gencove_url_vcr
from gencove.tests.utils import get_vcr_response

import pytest

from vcr import VCR


@pytest.fixture(scope="module")
def vcr_config():
    """VCR configuration."""
    return {
        "cassette_library_dir": "gencove/tests/files/vcr",
        "filter_headers": [
            "Authorization",
            "Content-Length",
            "User-Agent",
        ],
        "filter_post_data_parameters": [
            ("email", "email@example.com"),
            ("password", "mock_password"),
        ],
        "match_on": ["method", "scheme", "port", "path", "query"],
        "path_transformer": VCR.ensure_suffix(".yaml"),
        "before_record_request": [
            replace_gencove_url_vcr,
            filter_file_types_request,
        ],
        "before_record_response": [
            filter_jwt,
        ],
    }


@pytest.mark.vcr
@assert_authorization
def test_list_empty(mocker, credentials):
    """Test empty file types list"""
    runner = CliRunner()
    mocked_get_file_types = mocker.patch.object(
        APIClient,
        "get_file_types",
        return_value=FileTypesModel(results=[], meta=dict(next=None)),
    )
    res = runner.invoke(list_file_types, credentials)
    assert res.exit_code == 0
    mocked_get_file_types.assert_called_once()
    assert "" in res.output


@pytest.mark.vcr
@assert_authorization
def test_list_file_types_bad_project_id(
    mocker, credentials
):  # pylint: disable=unused-argument
    """Test file types throw an error if project is not uuid."""
    runner = CliRunner()

    project_id = "test"
    res = runner.invoke(list_file_types, ["--project-id", project_id, *credentials])
    assert res.exit_code == 1
    assert "ERROR: Project ID is not valid. Exiting." in res.output


@pytest.mark.vcr
@assert_authorization
def test_list_file_types_project_doesnt_exist(mocker, credentials, recording, vcr):
    """Test file types throw an error if no project doesn't exist."""
    runner = CliRunner()
    if not recording:
        file_types_response = get_vcr_response(
            "/api/v2/file-types/",
            vcr,
            operator.contains,
            just_body=False,
        )
        mocked_get_file_types = mocker.patch.object(
            APIClient,
            "get_file_types",
            side_effect=APIClientError(
                message="API Client Error: Not Found: Not found.",
                status_code=file_types_response["status"]["code"],
            ),
            return_value=file_types_response["body"]["string"],
        )

    project_id = str(uuid4())

    res = runner.invoke(
        list_file_types,
        ["--project-id", project_id, *credentials],
    )
    assert res.exit_code == 1
    assert (
        "\n".join(
            [
                f"ERROR: Project {project_id} does not exist.",
                "ERROR: API Client Error: Not Found: Not found.",
                "Aborted!\n",
            ]
        )
        == res.output
    )
    if not recording:
        mocked_get_file_types.assert_called_once()


@pytest.mark.vcr
@assert_authorization
@pytest.mark.parametrize("object_param", [None, "sample"])
def test_list_file_types_no_project(mocker, credentials, recording, vcr, object_param):
    """Test file types returns all sample file types if no project is given."""
    runner = CliRunner()
    if not recording:
        # Mock list_file_types only if using cassettes since return value is mocked.
        file_types_response = get_vcr_response(
            "/api/v2/file-types/", vcr, operator.contains
        )
        mocked_get_file_types = mocker.patch.object(
            APIClient,
            "get_file_types",
            return_value=FileTypesModel(**file_types_response),
        )
    args = [*credentials]
    if object_param is not None:
        args.extend(["--object", object_param])
    res = runner.invoke(list_file_types, [*args])
    assert res.exit_code == 0
    if not recording:
        mocked_get_file_types.assert_called_once()
        file_types = file_types_response["results"]
        file_types = "\n".join(
            [
                "\t".join(
                    [
                        file_type["key"],
                        file_type["description"],
                    ]
                )
                for file_type in sorted(file_types, key=lambda ft: ft["key"])
            ]
        )
        assert f"{file_types}\n" == res.output


@pytest.mark.vcr
@assert_authorization
@pytest.mark.parametrize("object_param", [None, "sample"])
def test_list_file_types_by_project(
    mocker, credentials, project_id, recording, vcr, object_param
):
    """Test file types being outputted to the shell"""
    runner = CliRunner()

    if not recording:
        # Mock list_file_types only if using cassettes since return value is mocked.
        file_types_response = get_vcr_response(
            "/api/v2/file-types/", vcr, operator.contains
        )
        mocked_get_file_types = mocker.patch.object(
            APIClient,
            "get_file_types",
            return_value=FileTypesModel(**file_types_response),
        )
    args = ["--project-id", project_id, *credentials]
    if object_param is not None:
        args.extend(["--object", object_param])
    res = runner.invoke(list_file_types, [*args])

    assert res.exit_code == 0

    if not recording:
        mocked_get_file_types.assert_called_once()
        file_types = file_types_response["results"]
        file_types = "\n".join(
            [
                "\t".join(
                    [
                        file_type["key"],
                        file_type["description"],
                    ]
                )
                for file_type in sorted(file_types, key=lambda ft: ft["key"])
            ]
        )
        assert f"{file_types}\n" == res.output


@pytest.mark.vcr
@assert_authorization
def test_list_file_types_no_project_reference_genome(
    mocker, credentials, recording, vcr
):
    """Test file types returns all reference genome types if no project is given."""
    runner = CliRunner()
    if not recording:
        # Mock list_file_types only if using cassettes since return value is mocked.
        file_types_response = get_vcr_response(
            "/api/v2/file-types/", vcr, operator.contains
        )
        mocked_get_file_types = mocker.patch.object(
            APIClient,
            "get_file_types",
            return_value=FileTypesModel(**file_types_response),
        )
    res = runner.invoke(list_file_types, ["--object", "reference_genome", *credentials])
    assert res.exit_code == 0
    if not recording:
        mocked_get_file_types.assert_called_once_with(object_param="reference_genome")
        file_types = file_types_response["results"]
        file_types = "\n".join(
            [
                "\t".join(
                    [
                        file_type["key"],
                        file_type["description"],
                    ]
                )
                for file_type in sorted(file_types, key=lambda ft: ft["key"])
            ]
        )
        assert f"{file_types}\n" == res.output


@pytest.mark.vcr
@assert_authorization
def test_list_file_types_by_project_reference_genome(
    mocker, credentials, project_id, recording, vcr
):
    """Test that no ouput is given if Project doesn't have a Reference Genome."""
    runner = CliRunner()
    if not recording:
        # Mock list_file_types only if using cassettes since return value is mocked.
        file_types_response = get_vcr_response(
            "/api/v2/file-types/", vcr, operator.contains
        )
        mocked_get_file_types = mocker.patch.object(
            APIClient,
            "get_file_types",
            return_value=FileTypesModel(**file_types_response),
        )
    res = runner.invoke(
        list_file_types,
        ["--object", "reference_genome", "--project-id", project_id, *credentials],
    )

    assert res.exit_code == 0
    if not recording:
        mocked_get_file_types.assert_called_once_with(
            project_id=project_id, object_param="reference_genome"
        )
        file_types = file_types_response["results"]
        file_types = "\n".join(
            [
                "\t".join(
                    [
                        file_type["key"],
                        file_type["description"],
                    ]
                )
                for file_type in sorted(file_types, key=lambda ft: ft["key"])
            ]
        )
        assert f"{file_types}" == res.output



"""Project create shell command definition."""
import click

from gencove.command.common_cli_options import add_options, common_options
from gencove.command.utils import validate_uuid
from gencove.constants import Credentials, Optionals

from .main import CreateProject


@click.command("create")
@click.argument("project_name")
@click.argument("pipeline_capability_id", callback=validate_uuid)
@add_options(common_options)
def create_project(  # pylint: disable=too-many-arguments
    project_name,
    pipeline_capability_id,
    host,
    email,
    password,
    api_key,
):
    """Create a project.

    `PROJECT_NAME`: name of the Gencove project

    `PIPELINE_CAPABILITY_ID`: id of the pipeline capability to be used for the project
    """

    CreateProject(
        project_name,
        pipeline_capability_id,
        Credentials(email=email, password=password, api_key=api_key),
        Optionals(host=host),
    ).run()



# This file needs to be called conftest.py. See:
# https://docs.pytest.org/en/6.2.x/fixture.html#conftest-py-sharing-fixtures-across-multiple-files
"""Common fixtures."""

# pylint: disable=import-error

import os
import uuid
from datetime import datetime

import pytest


@pytest.fixture(scope="session")
def using_api_key():
    """Returns True if API Key is being used."""
    use_api_key = os.getenv("USING_API_KEY")
    return use_api_key is not None


@pytest.fixture(scope="session")
def recording(record_mode):
    """Returns True if VCR cassettes are being recorded.
    False if being played.
    """
    return record_mode != "none"


@pytest.fixture(scope="session")
def project_id():
    """Returns the project id."""
    return os.getenv("GENCOVE_PROJECT_ID")


@pytest.fixture(scope="session")
def pipeline_capability_id():
    """Returns the pipeline capability id."""
    return os.getenv("GENCOVE_PIPELINE_CAPABILITY_ID")


@pytest.fixture(scope="session")
def pipeline_id():
    """Returns the pipeline id."""
    return os.getenv("GENCOVE_PIPELINE_ID")


@pytest.fixture(scope="session")
def project_id_batches():
    """Returns the project id that contains batches."""
    return os.getenv("GENCOVE_PROJECT_ID_BATCHES_TEST")


@pytest.fixture(scope="session")
def project_id_sample_manifest():
    """Returns the project id that contains batches."""
    return os.getenv("GENCOVE_PROJECT_ID_SAMPLE_MANIFEST_TEST")


@pytest.fixture(scope="session")
def project_id_download():
    """Returns the project id that contains batches."""
    return os.getenv("GENCOVE_PROJECT_ID_DOWNLOAD_TEST")


@pytest.fixture(scope="session")
def project_id_delete():
    """Returns the project id to be deleted."""
    return os.getenv("GENCOVE_PROJECT_ID_DELETE_TEST")


@pytest.fixture(scope="session")
def project_id_reference_genome():
    """Returns the project id to be deleted."""
    return os.getenv("GENCOVE_PROJECT_ID_REFERENCE_GENOME_TEST")


@pytest.fixture(scope="session")
def project_id_import():
    """Returns the project id to import into."""
    return os.getenv("GENCOVE_PROJECT_ID_IMPORT_TEST")


@pytest.fixture(scope="session")
def credentials(using_api_key):  # pylint: disable=redefined-outer-name
    """Fixture to have the appropriate credentials."""
    api_key = os.getenv("GENCOVE_API_KEY_TEST")
    email = os.getenv("GENCOVE_EMAIL_TEST")
    password = os.getenv("GENCOVE_PASSWORD_TEST")
    if using_api_key:
        return ["--api-key", api_key]
    return ["--email", email, "--password", password]


@pytest.fixture(scope="session")
def archived_sample():
    """Returns the archived sample id.."""
    return os.getenv("GENCOVE_ARCHIVED_SAMPLE_TEST")


@pytest.fixture(scope="session")
def deleted_sample():
    """Returns deleted sample id"""
    return os.getenv("GENCOVE_DELETED_SAMPLE_TEST")


@pytest.fixture(scope="session")
def sample_manifest_id():
    """Returns a sample manifest id."""
    return os.getenv("GENCOVE_SAMPLE_MANIFEST_ID_TEST")


@pytest.fixture(scope="session")
def sample_id_batches():
    """Returns the sample id to create batches."""
    return os.getenv("GENCOVE_SAMPLE_ID_BATCHES_TEST")


@pytest.fixture(scope="session")
def sample_id_download():
    """Returns the sample id to download."""
    return os.getenv("GENCOVE_SAMPLE_ID_DOWNLOAD_TEST")


@pytest.fixture(scope="session")
def batch_type():
    """Returns the batch type to create batches."""
    return os.getenv("GENCOVE_BATCH_TYPE_TEST")


@pytest.fixture(scope="session")
def batch_name():
    """Returns the batch name to create batches."""
    return os.getenv("GENCOVE_BATCH_NAME_TEST")


@pytest.fixture(scope="session")
def batch_id():
    """Returns a batch id."""
    return os.getenv("GENCOVE_BATCH_ID_TEST")


@pytest.fixture(scope="session")
def sample_id_import_existing():
    """Returns a successful sample id."""
    return os.getenv("GENCOVE_SAMPLE_ID_IMPORT_EXISTING_TEST")


@pytest.fixture(scope="function", autouse=True)
def dont_save_dump_log():
    """Sets the environment variable to disable dumping the log file."""
    os.environ["GENCOVE_SAVE_DUMP_LOG"] = "FALSE"


@pytest.fixture(scope="function")
def dump_filename(mocker):
    """Fixtures that returns the log filename and creates the folder."""
    random_id = str(uuid.uuid4())
    now = datetime.utcnow()
    folder = f".logs/{now:%Y_%m}"
    filename = f"{folder}/{now:%Y_%m_%d_%H_%M_%S}_{random_id[:8]}.log"

    def get_debug_file_name_patch():
        os.makedirs(folder)
        return filename

    mocker.patch(
        "gencove.logger.get_debug_file_name",
        side_effect=get_debug_file_name_patch,
    )
    return filename


@pytest.fixture(scope="function")
def sample_archive_status_null():
    """A sample containing a NULL (eg. deleted) archive status"""
    return {
        "meta": {"count": 1, "next": None, "previous": None},
        "results": [
            {
                "id": "11111111-1111-1111-1111-111111111111",
                "created": "2021-09-21T18:30:44.799519Z",
                "modified": "2021-09-21T18:38:57.735776Z",
                "client_id": "mock client_id",
                "physical_id": "",
                "legacy_id": "",
                "last_status": {
                    "id": "11111111-1111-1111-1111-111111111111",
                    "status": "mock status",
                    "note": "",
                    "created": "2021-09-21T18:38:57.735776Z",
                },
                "archive_last_status": None,
            }
        ],
    }



"""Configure inactivity stop for explorer instances subcommand."""

from .utils import calculate_applied_hours_to_instance, hours_to_human_readable
from ....base import Command
from .....exceptions import ValidationError


class StopInstanceInactivity(Command):
    """Configure inactivity stop for explorer instances command executor."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        hours,
        organization,
        override,
        credentials,
        options,
    ):
        super().__init__(credentials, options)
        self.hours = hours
        self.organization = organization
        self.override = override
        self.hours_is_empty = False

    def initialize(self):
        """Initialize inactivity-stop subcommand."""
        self.login()

    def validate(self):
        """Validate command input."""
        self.echo_debug("Validating instance-inactivity command input")

        self.hours_is_empty = self.hours is None
        if self.hours in ["None", "none", "null"]:
            self.hours = None
        if not self.hours_is_empty and self.hours is not None:
            self.hours = int(self.hours)
        if not self.hours_is_empty and self.hours is not None and self.hours < 0:
            raise ValidationError("Expected 'hours' to be a positive integer or None.")
        if self.organization and self.hours is None:
            raise ValidationError(
                "Expected 'hours' to be a positive integer if '--organization' is provided."  # noqa: E501, pylint: disable=line-too-long
            )
        if self.override is not None and not self.organization:
            raise ValidationError(
                "Expected '--override' to be used with '--organization' flag."
            )

    def execute(self):
        """Make a request to configure instance stop inactivity."""
        self.echo_debug("Configure explorer instances stop after inactivity.")

        instances_config, org_config = (
            self.api_client.get_explorer_instances().dict(),
            self.api_client.get_explorer_instances_activity_stop_organization().dict(),
        )

        if self.organization:
            self.echo_debug("Reading organization config")
            request_data = {}
            if self.override is not None:
                request_data[
                    "explorer_override_stop_after_inactivity_hours"
                ] = self.override
            if not self.hours_is_empty:
                request_data["explorer_stop_after_inactivity_hours"] = self.hours
            if request_data:
                self.echo_debug("Setting configuration organization wide")
                org_config.update(request_data)
                self.api_client.set_explorer_instances_activity_stop_organization(
                    org_config["explorer_override_stop_after_inactivity_hours"],
                    org_config["explorer_stop_after_inactivity_hours"],
                )
        else:
            self.echo_debug("Reading instances config")
            if not self.hours_is_empty:
                self.echo_debug("Setting configuration for instances")
                self.api_client.set_explorer_instances_activity_stop(
                    [e["id"] for e in instances_config["results"]], self.hours
                )
                instances_config["results"] = [
                    {"id": e["id"], "stop_after_inactivity_hours": self.hours}
                    for e in instances_config["results"]
                ]

        self.show_inactivity_config(instances_config, org_config)

    def show_inactivity_config(self, instances_config, org_config):
        """Display inactivity config"""
        self.echo_debug("Displaying inactivity config")
        self.echo_info("Inactivity stop configuration")
        if self.organization:
            org_hours = hours_to_human_readable(
                org_config["explorer_stop_after_inactivity_hours"]
            )
            org_override = org_config["explorer_override_stop_after_inactivity_hours"]
            self.echo_info(
                f"Organization:\t\t\t\t\thours={org_hours}, override={org_override}"
            )
        else:
            for instance in instances_config["results"]:
                instance_id = str(instance["id"]).replace("-", "")
                instance_hours = hours_to_human_readable(
                    instance["stop_after_inactivity_hours"]
                )
                applied_hours, from_org = calculate_applied_hours_to_instance(
                    instance, org_config
                )
                if from_org:
                    original_config = f", instance_config[hours={instance_hours}]"
                else:
                    original_config = ""
                applied_hours = hours_to_human_readable(
                    applied_hours, from_org=from_org
                )
                self.echo_info(
                    f"Instance {instance_id}:\thours={applied_hours}{original_config}"
                )
